require 'rest_client'
require 'net/http'
require 'json'
require 'hashie'
require 'asm/util'
require 'shellwords'

module ASM
  module Facts
    class ReplaceFactsError < StandardError;
    end
    class FactScriptError < StandardError;
    end

    def self.run_script(device_config, options = {})
      options[:logger] ||= ASM.logger
      options[:cache_file] ||= "/opt/Dell/ASM/cache/#{device_config[:cert_name]}.json"

      path = device_config[:path]
      script_path = "#{ASM::Util::DEVICE_MODULE_PATH}#{path}"
      if !File.exists?(script_path) || !File.executable?(script_path)
        raise(ASM::DeviceManagement::FactRetrieveError,
              "Unable to retrieve facts for %s, no executable script to run" % device_config[:provider])
      end

      facts = {
          'device_type' => 'script',
          'certname' => device_config[:cert_name]
      }

      begin
        args = build_command(device_config)

        # TODO: The intent was for inventory scripts that support --file to save
        # their json output at the file location specified by that argument. That
        # would allow the script to output log messages that could be used for
        # debugging without breaking the format of the final JSON
        # results. However the scripts currently implementing this option a)
        # don't log any debug output and b) save their JSON output at a
        # hard-coded location, /opt/Dell/ASM/cache/[IP].json. The hard-coded
        # location is problematic for recent queued inventory which
        # expects to be able to save the cache file output in specific job
        # directories. For new asm-deployer never sets the --file option.
        has_file_option = can_write_to_file?(script_path)
        has_file_option = false

        if has_file_option
          args << "--file"
          args << options[:cache_file]
        end
        masked_args = [script_path] + args.dup
        masked_args[masked_args.find_index {|x| x == '--password'} + 1] = '*******'
        masked_cmd = masked_args.join(' ')
        password_index = args.find_index { |x| x == "--password" }
        password = args[password_index + 1]
        # Password as environment variable
        args.slice!(password_index .. password_index + 1)
        options[:logger].info("Executing inventory command: %s" % masked_cmd)

        result = nil
        # Would prefer to use run_command_streaming here but that
        # hits a jruby 1.7 bug where back-slashes in the command get doubled,
        # which plays havoc with domain\username arguments. Using
        # run_command_with args and printing all stderr last instead.
        result = ASM::Util.run_with_clean_env(script_path, false, *args, 'PASSWORD'.to_sym => password)

        unless result && result.exit_status == 0
          raise(FactScriptError, "Inventory command returned exit code #{result.exit_status}\n #{result.stdout}")
        end

        if options[:log_file]
          File.open(options[:log_file], 'w') do |f|
            f.puts result.stdout
            f.puts result.stderr
          end
        end

        if has_file_option
          raise(ASM::DeviceManagement::FactRetrieveError,
                "Error, missing fact discovery file %s" % options[:cache_file]) unless File.exists? options[:cache_file]
          data = JSON.parse(File.read(options[:cache_file]))
        else
          data = JSON.parse(result.stdout)
        end

        # until we are on the new puppetdb if we detect we are getting structured data
        # store the data as a JSON encoded string into the 'json_facts' fact
        if data.reject { |k, v| String == v.class }.empty?
          facts.merge!(data)
        else
          facts.merge!({'json_facts' => JSON.dump(data)})
        end
      end
      facts
    end

    def self.build_command(device_config)
      args = []
      [[:user, 'username'], :password, :port, [:host, 'server']].each do |k, name|
        if device_config[k]
          name ||= k.to_s
          args << "--#{name}"
          args << device_config[k].to_s
        end
      end

      # Add all device_config[:arguments] except credential_id. Everything it
      # contains (username, password, community_string) is already passed as
      # separate plaintext arguments
      device_config[:arguments].find_all { |k, _| k != "credential_id" }.each do |k, v|
        args << "--%s" % k.gsub('_', '-')
        args << v
      end
      args
    end

    def self.can_write_to_file?(script_path)
      `#{script_path} --help`.include? '--file'
    end

  end
end
