require 'asm/util'

module ASM
  module Ipmi

    class Error < StandardError; end

    # Wrapper for the IPMI CLI commands. endpoint should be a hash of
    # :host, :user, :password
    def self.invoke(endpoint, command, options = {})
      options = {
          :selector => nil,
          :props => {},
          :logger => nil,
          :nth_attempt => 0,
      }.merge(options)

      unless options[:logger].nil? || options[:logger].respond_to?(:error)
        # The Puppet class has most of the methods loggers respond to except for error
        logger = options[:logger]
        def logger.error(msg)
          self.err(msg)
        end
      end

      base_cmd = '/usr/bin/ipmitool'
      args = [ '-I', 'lanplus', '-H', endpoint[:host],
                '-U', endpoint[:user], '-P', endpoint[:password], command]
      base_cmd = base_cmd + ' ' + args.join(' ')
      if options[:logger]
        masked_args = args.dup
        masked_args[args.find_index('-P') + 1] = '******'
        options[:logger].debug("Executing /usr/bin/ipmitool #{masked_args.join(' ')}")
      end
      result = ASM::Util.run_command_with_args(base_cmd)
      options[:logger].debug("Result = #{result}") if options[:logger]

      unless result.exit_status == 0 && result.stderr.empty?
        if result['stdout'] =~ /Unable to establish IPMI/
          if options[:nth_attempt] < 2
            options[:nth_attempt] += 1
            options[:logger].info("Unable to access IPMI interface, retrying #{endpoint[:host]}...")
            return invoke(endpoint, command, options)
          end
          msg = "Unable to establish IPMI, please retry with correct credentials at #{endpoint[:host]}."
        else
          msg = "Failed to execute IPMI command against server #{endpoint[:host]}"
        end
        options[:logger].error(msg) if options[:logger]
        raise(Error, "#{msg}: #{result}")
      end

      result['stdout']
    end

    def self.reboot(endpoint, logger = nil)
      if self.get_power_status(endpoint,logger) == 'off'
        logger.info('Server is powered-off. Need to power-on the server') if logger
        command = 'power on'
        self.invoke(endpoint, command, options = {:logger => logger})
      else
        command = 'power cycle'
        self.invoke(endpoint, command, options = {:logger => logger})
      end
    end

    def self.get_power_status(endpoint, logger = nil)
      command = 'power status'
      power_status = self.invoke(endpoint, command, options = {:logger => logger})
      power_status = power_status.scan(/Chassis Power is\s+(\S+)$/m).flatten.first.strip
      logger.debug("Current power status: #{power_status}") if logger
      power_status
    end

    def self.power_on(endpoint, logger = nil)
      if self.get_power_status(endpoint,logger) == 'on'
        logger.info("Server is already powered-on.") if logger
        return true
      end
      command = 'power on'
      power_on_op = self.invoke(endpoint, command, options = {:logger => logger})
    end

    def self.power_off(endpoint, logger = nil)
      if self.get_power_status(endpoint,logger) == 'off'
        logger.info("Server is already powered-off.") if logger
        return true
      end
      command = 'power off'
      self.invoke(endpoint, command, options = {:logger => logger})
    end

    def self.sleep_time
      60
    end

  end
end
