module ASM
  class Provider
    class Cluster
      class Vmware < Provider::Base
        puppet_type "asm::cluster"

        property    :datacenter,      :default => nil,                        :validate => String
        property    :cluster,         :default => nil,                        :validate => String
        property    :vcenter_options, :default => {"insecure" => true},       :validate => Hash
        property    :ensure,          :default => "present",                  :validate => ["present", "absent"]

        def should_inventory?
          !!type.guid
        end

        def update_inventory
          raise("Cannot update inventory for %s without a guid" % type.puppet_certname) unless type.guid

          ASM::Util.update_asm_inventory(type.guid) unless debug?
        end

        def virtualmachine_supported?(vm)
          vm.provider_path == "virtualmachine/vmware"
        end

        def evict_server!(server)
          hostname = type.deployment.lookup_hostname(server.hostname, server.static_network_config).downcase
          puppet_hash = asm_host_hash(server.puppet_certname, hostname, host_username, server.admin_password, "absent", true)

          logger.debug("Removing server %s from the cluster %s" % [server.puppet_certname, type.puppet_certname])

          type.process_generic(server.puppet_certname, puppet_hash, puppet_run_type, true)
        end

        # Removes a volume from a VMWare cluster, these volumes need to be removed from
        # every server and the data store resource is specific to the type of storage so
        # using the Type::Volume#esx_datastore helper it construct the correct esx_datastore
        # resource for every server and then run them all together
        #
        # Servers that are being torndown is skipped as they'd already be gone or going away
        # in the same deployment so there's no need
        def evict_volume!(volume)
          resources = {}

          type.related_servers.each do |server|
            if !server.teardown?
              logger.debug("Removing volume %s from server %s for cluster %s" % [volume.puppet_certname, server.puppet_certname, type.puppet_certname])
              resources.merge!(volume.esx_datastore(server, type, "absent"))
            end
          end

          unless resources.empty?
            resources.merge!(transport_config)
            type.process_generic(type.puppet_certname, resources, puppet_run_type, true, nil, type.guid)
          end
        end

        # Generates a transport configuration used by Puppet
        #
        # @param name [String] The name of the vcenter instance
        # @return [Hash] a transport structure, usually combined with other data
        def transport_config(name=nil)
          name ||= type.puppet_certname

          {"transport" =>
            {"vcenter" =>
              {"name" => name,
               "options" => {"insecure" => true},
               "provider" => "device_file"
              }
            }
          }
        end


        def host_username
          ASM::ServiceDeployment::ESXI_ADMIN_USER
        end

        def asm_host_hash(certname, hostname, username, password, host_ensure, include_transport=false)
          h = {"asm::host" =>
                {certname =>
                  {'datacenter' => datacenter,
                   'cluster' => cluster,
                   'hostname' => hostname,
                   'username' => username,
                   'password' => password,
                   'decrypt' => @type.deployment.decrypt?,
                   'timeout' => 90,
                   'ensure' => host_ensure
                  }
                }
              }

          include_transport ? h.merge(transport_config) : h
        end

        def transport_config(name=uuid)
          {'transport' =>
            {'vcenter' =>
              {'name' => name,
               'options' => {'insecure' => true},
               'provider' => 'device_file'
              }
            }
          }
        end
      end
    end
  end
end
