require 'asm/item_validator'

module ASM
  class Provider
    # helper module to provider hash like access to provider properties but with
    # built in concepts of default values and validation.
    #
    # code is based on https://github.com/ripienaar/gwtf/blob/master/lib/objhash.rb
    module Phash
      include Enumerable

      module ClassMethods
        # Defines a property including it's defaults and validation
        #
        # @macro [attach] phash.property
        #   Provides access to the $1 property of the resource
        #
        #   When setting a value using $1=(value) validation is done using {ASM::ItemValidator}
        #   and inputs are expected to pass validation as defined in +$3+
        #
        #   When accessing a value using the $1 method the default is returned when nothing
        #   has been set, defaults can be something that do not pass validation - something can
        #   be nil by default while setting something would require a String matching a regex
        #   for example
        #
        #   @method $1
        #   @method $1=(value)
        #
        #   @attribute
        #   @return The value stored, +$2+ if none is set, type depends on stored data
        #   @raise [StandardError] when attempting read or write to a non existing property
        #   @raise [StandardError] when attempting to write data that does not pass validation
        def property(name, args={:default => nil, :validation => nil, :tag => [:puppet]})
          name = name.to_s

          raise("Already have a property %s" % name) if phash_config.include?(name)

          args[:tag] = Array(args[:tag]) if args[:tag]

          phash_config[name] = {:default => nil, :validation => nil, :tag => [:puppet]}.merge(args)
        end

        # @api private
        def phash_config
          @__phash_config ||= { }
        end

        # @api private
        def phash_default_value(property)
          property = property.to_s

          raise("Unknown property %s" % property) unless phash_config.include?(property)

          if phash_config[property][:default].is_a?(Proc)
            phash_config[property][:default].call
          else
            phash_config[property][:default]
          end
        end

        # @api private
        def phash_reset!
          @__phash_config = {}
        end
      end

      # @api private
      def self.included(base)
        base.extend ClassMethods
      end

      # @api private
      def default_property_value(property)
        default = self.class.phash_default_value(property)

        begin
          default = Marshal.load(Marshal.dump(default))
        rescue
        end

        default
      end

      # @api private
      def phash_config
        self.class.phash_config
      end

      # @api private
      def phash_values
        return @__phash_values if @__phash_values

        @__phash_values = {}

        phash_config.each_pair do |property, args|
          update_property(property, default_property_value(property))
        end

        @__phash_values
      end

      def include?(property)
        phash_config.include?(property.to_s)
      end

      # @api private
      def update_property(property, value)
        property = property.to_s

        raise("Unknown property %s" % property) unless include?(property)

        validate_property(property, value)

        phash_values[property] = value
      end

      # @api private
      def validate_property(property, value)
        raise("Unknown property %s" % property) unless include?(property)

        validation = phash_config[property.to_s][:validation]

        return true if validation.nil?

        # if the value is the default we dont validate it allowing nil
        # defaults but validation only on assignment of non default value
        return true if value == phash_config[property.to_s][:default]

        raise("%s should be %s but got %s" % [property, validation, value.inspect]) if value.nil? && !validation.nil?

        validated, fail_msg = ItemValidator.validate!(value, validation)

        return true if validated

        raise("%s failed to validate: %s" % [property.to_s, fail_msg])
      end

      def to_hash(exclude_nil=false, tagged=[])
        hash = {}

        properties(tagged).each do |property|
          if !(exclude_nil && self[property].nil?)
            hash[property] = self[property]
          end
        end

        hash
      end

      def each
        properties.sort.each do |property|
          yield [property, phash_values[property]]
        end
      end

      def [](property)
        raise("No such property: %s" % property) unless include?(property)

        phash_values[property.to_s]
      end

      def []=(property, value)
        update_property(property, value)
      end

      def merge(hsh)
        raise(TypeError, "Can't merge %s into Hash" % hsh.class) unless hsh.respond_to?(:to_hash)

        phash_values.merge(hsh)
      end

      def merge!(hsh)
        raise(TypeError, "Can't merge %s into Hash" % hsh.class) unless hsh.respond_to?(:to_hash)

        properties.each do |k|
          update_property(k, hsh[k]) if hsh.include?(k)
        end

        self
      end

      # Retrieve a sorted list of known properties regardless of their value
      #
      # When supplying the tagged argument it should be a list of tags, any
      # property that has at least all these tags - or no tags - will be included
      # in the list.  Keep in mind the default tag for any unspecified tag when
      # creating a property is [:puppet]
      #
      # @param tagged [Array] Retrieve properties with all these tags
      # @return [Array] property names
      def properties(tagged=nil)
        if tagged
          tagged = Array(tagged)

          phash_values.select do |k, v|
            (tagged - phash_config[k][:tag]).empty?
          end.keys.sort
        else
          phash_values.keys.map{|k| k.to_s}.sort
        end
      end

      # Provides access to the hash using a object like methods
      #
      # @example simple read from the class
      #
      #   i.description #=> "Sample Item"
      #
      # @example method like writes
      #
      #   i.description "This is a test" #=> "This is a test"
      #
      # @example assignment
      #
      #   i.description = "This is a test" #=> "This is a test"
      #
      # @raise [NameError] for unknown keys or access to methods not shown above
      def method_missing(method, *args)
        method = method.to_s

        if include?(method)
          if args.empty?
            return phash_values[method]
          else
            return update_property(method, args.first)
          end

        elsif method =~ /^(.+)=$/
          property = $1
          return update_property(property, args.first) if include?(property)
        end

        raise(NameError, "undefined local variable or method `%s'" % method)
      end
    end
  end
end

