require 'asm/network_configuration'
require 'asm/wsman'
require 'asm/ipmi'

module ASM
  class Provider
    class Server
      class Server < Provider::Base
        puppet_type "asm::server"

        property  :razor_image,              :default => nil,        :validation => String
        property  :admin_password,           :default => nil,        :validation => String
        property  :os_host_name,             :default => nil,        :validation => String
        property  :serial_number,            :default => nil,        :validation => String
        property  :policy_name,              :default => nil,        :validation => String
        property  :os_image_type,            :default => nil,        :validation => String
        property  :os_image_version,         :default => nil,        :validation => String
        property  :broker_type,              :default => "noop",     :validation => String
        property  :ensure,                   :default => "present",  :validation => ["present", "absent"]
        property  :decrypt,                  :default => false,      :validation => :boolean
        property  :installer_options,        :default => {},         :validation => Hash
        property  :esx_mem,                  :default => "",         :validation => String
        property  :razor_api_options,        :default => nil,        :validation => Hash
        property  :client_cert,              :default => nil,        :validation => String

        property  :domain_admin_user,        :default => nil,        :validate => String,  :tag => :hyperv
        property  :domain_admin_password,    :default => nil,        :validate => String,  :tag => :hyperv
        property  :domain_name,              :default => nil,        :validate => String,  :tag => :hyperv
        property  :fqdn,                     :default => nil,        :validate => String,  :tag => :hyperv

        property  :ntp_server,               :default => nil,        :validate => String,  :tag => :installer_options
        property  :os_type,                  :default => nil,        :validate => String,  :tag => :installer_options
        property  :language,                 :default => nil,        :validate => String,  :tag => :installer_options
        property  :keyboard,                 :default => nil,        :validate => String,  :tag => :installer_options
        property  :product_key,              :default => nil,        :validate => String,  :tag => :installer_options
        property  :time_zone,                :default => nil,        :validate => String,  :tag => :installer_options # linux
        property  :timezone,                 :default => nil,        :validate => String,  :tag => :installer_options # hyperv

        BOOT_FROM_SAN_TARGETS = ["FC", "iSCSI"]

        # The deployment includes component resources that are not strictly resources
        # but are instead used to convey information for the rest of the deployment,
        # so a special case to_puppet here will remove those when they exist leaving
        # only asm::server
        def to_puppet
          Hash[super.select do |key|
            key == "asm::server"
          end]
        end

        def configure_hook
          self[:serial_number] ||= type.cert2serial
          self[:policy_name] ||= type.razor_policy_name
          self[:razor_api_options] ||= ASM.config.http_client_options || {}
          self[:razor_api_options]["url"] ||= ASM.config.url.razor
          self[:client_cert] ||= ASM.config.client_cert

          configure_installer_options
        end

        def should_inventory?
          !!type.guid
        end

        def update_inventory
          raise("Cannot update inventory for %s without a guid" % type.puppet_certname) unless type.guid

          ASM::Util.update_asm_inventory(type.guid) unless debug?
        end

        # (see Type::Server#physical_type)
        def physical_type
          inventory = type.retrieve_inventory
          type = inventory["serverType"]

          type = "SLED" if inventory["model"].match(/PowerEdge FC\d+/)

          type
        end

        def controller_supported?(controller)
          controller.provider_path == "controller/idrac"
        end

        def cluster_supported?(cluster)
          case cluster.provider_path
          when "cluster/scvmm"
            configured_for_hyperv?
          else
            true
          end
        end

        def configured_for_hyperv?
          hyperv_properties = to_hash(true, :hyperv).keys

          missing = (Provider::Cluster::Scvmm::SERVER_REQUIRED_PROPERTIES - hyperv_properties)

          return true if missing.empty?

          logger.warn("Server %s is not supported by HyperV as it lacks these properties or they have nil values: %s" % [type.puppet_certname, missing.join(", ")])
          false
        end

        def configure_installer_options
          properties(:installer_options).each do |option|
            if self[option]
              self[:installer_options][option] = self[option]
            end
          end

          self[:installer_options]["os_type"] = self[:os_image_type]
          self[:installer_options]["agent_certname"] = hostname_to_certname if hostname
        end

        def hostname
          self[:os_host_name]
        end

        def wait_for_lc_ready
          return if debug?

          if dell_server?
            ASM::WsMan.wait_for_lc_ready(type.device_config, logger)
          end
        end

        def hostname_to_certname(name=nil)
          name ||= hostname
          ASM::Util.hostname_to_certname(name)
        end

        def dell_server?(cert_name=nil)
          Util.dell_cert?(cert_name || type.puppet_certname)
        end

        # Look up the target_boot_device from the asm::idrac component resource
        #
        # @return [String, nil] the boot device or nil
        def target_boot_device
          idrac = type.service_component.resource_by_id("asm::idrac")

          idrac ? idrac["target_boot_device"] : nil
        end

        def boot_from_san?
          return false unless dell_server?

          BOOT_FROM_SAN_TARGETS.include?(target_boot_device)
        end

        def has_os?
          !os_image_type.nil?
        end

        def idrac?
          type.service_component.has_resource_id?("asm::idrac")
        end

        # Creates a Type::Controller instance from the asm::idrac component resource
        #
        # @return [Type::Controller, nil] nil when it could not be created
        def create_idrac_resource
          idrac = nil

          if resource = type.service_component.resource_by_id("asm::idrac")
            component = resource.to_component(type.puppet_certname, "CONTROLLER")
            idrac = component.to_resource(type.deployment, logger)

            if idrac.supports_resource?(type)
              idrac.configure_for_server(type)
            else
              logger.warn("Created the asm::idrac resource but it does not support this server")
              idrac = nil
            end
          end

          idrac
        end

        def delete_server_cert!
          # without an os_image_type it would never have run puppet so would not have a cert
          return if os_image_type == "" || os_image_type.nil?

          logger.info("Deleting certificate for server %s" % os_host_name)
          ASM::DeviceManagement.clean_cert(hostname_to_certname(os_host_name)) unless debug?
        end

        def delete_server_node_data!
          return if os_image_type == "" || os_image_type.nil?

          logger.info("Deleting node data for server %s" % os_host_name)
          ASM::DeviceManagement.remove_node_data(hostname_to_certname(os_host_name)) unless debug?
        end

        def leave_cluster!
          cluster = type.related_cluster

          # We do not try and remove ourselves from clusters that are being torn down in the same
          # deployment as they would already be torn down at this point so this would fail anyway
          # even if they are not already torn down this would be a needless action and so this is
          # both avoiding errors in some cases and a performance optimisation in others
          if cluster
            if !cluster.teardown?
              cluster.evict_server!(type)
            else
              logger.debug("Server %s skipping cluster eviction as the %s cluster is also being torn down" % [type.puppet_certname, cluster.puppet_certname])
            end
          else
            logger.debug("Did not find any cluster associated with server %s" % type.puppet_certname)
          end
        end

        def clean_related_volumes!
          if idrac?
            type.related_volumes.each do |volume|
              # Skip volumes that are being torn down as the cleanup will happen
              # when they get torn down. This is a optimisation short circuit only
              # and not for functional reasons
              unless volume.teardown?
                logger.info("Removing associated storage access for volume %s after removal of server %s" % [volume.puppet_certname, type.puppet_certname])

                begin
                  volume.remove_server_from_volume!(type)
                rescue
                  logger.warn("Failed to remove the server %s from the volume %s: %s: %s" % [type.puppet_certname, volume.puppet_certname, $!.class, $!.to_s])
                  logger.debug($!.backtrace.inspect)
                end
              end
            end
          else
            logger.debug("Not attempting to clean any related volumes on non iDRAC server %s" % type.puppet_certname)
          end
        end

        def clean_virtual_identities!
          if idrac = create_idrac_resource
            logger.info("Cleaning virtual identities from %s" % type.puppet_certname)

            idrac.ensure = "teardown"
            idrac.process!

            wait_for_lc_ready
            else
              logger.warn("Server %s has no associated asm::idrac resource, cannot clean its virtual identities"  % type.puppet_certname)
            end
        end

        def power_off!
          if !debug?
            if dell_server?
              wait_for_lc_ready
              ASM::WsMan.poweroff(type.device_config, logger)
            else
              ASM::Ipmi.power_off(type.device_config, logger)
            end
          else
            logger.debug("Would have powered off %s but running in debug mode" % type.puppet_certname)
          end
        end
      end
    end
  end
end
