module ASM
  class RuleEngine
    class Rules
      attr_reader :rules

      # @param rules_dir [Array<string>] list of directories to look for rules
      # @param logger [Logger]
      def initialize(rules_dir, logger)
        @rules = []
        @logger = logger
        @rulesdir = Array(rules_dir)
      end

      # Look up a rule by name
      #
      # @param rule [String, Symbol] the rule name specified in {ASM::RuleEngine.new_rule}
      # @return [RuleEngine::Rule]
      def [](rule)
        @rules[rule]
      end

      # Amount of rules loaded
      #
      # @return [Fixnum] rule count
      def size
        @rules.size
      end

      # Add a rule to the collection
      #
      # @param rule [RuleEngine::Rule] rule to add
      # @return [RuleEngine::Rule]
      def add_rule(rule)
        @rules << rule
      end

      # yields each loaded rule in order of priority
      #
      # @yield [RuleEngine::Rule] each rule
      # @return [void]
      def by_priority
        @rules.sort_by{|r| r.priority}.each do |rule|
          yield(rule)
        end
      end

      # Load a rule from dis
      #
      # @param file [String] a file to read the rule from
      # @raise [StandardError] when the file cannot be found or read
      # @raise [StandardError] if rule creation fails for any reason
      # @return [void]
      def load_rule(file)
        raise("Cannot read file %s to load a rule from" % file) unless File.readable?(file)

        rule_body = File.read(file)

        cleanroom = Object.new
        rule = cleanroom.instance_eval(rule_body, file, 1)
        rule.file = file
        rule.logger = @logger

        add_rule(rule)

        @logger.debug("Loaded rule %s from %s" % [rule.name, file])
      end

      # Find correctly named rule files in a directory
      #
      # Valid rule files are named +something_rule.rb+
      #
      # @return [Array<String>] list of found rule file names without the directory part
      def find_rules(dir)
        if File.directory?(dir)
          Dir.entries(dir).grep(/_rule.rb$/)
        else
          @logger.debug("The argument %s is not a directory while looking for rules" % dir)
          []
        end
      end

      # Load all rules found in any of the directories specified in {#initialize}
      #
      # @return [void]
      def load_rules!
        @rulesdir.each do |dir|
          find_rules(dir).each do |rule|
            load_rule(File.join(dir, rule))
          end
        end
      end
    end
  end
end
