require 'asm/service/component'
require 'asm/service/component/resource'
require 'asm/service/component/generator'
require 'asm/service/processor'
require 'asm/service/rule_gen'

# wrapper class with helpers for the service template received from the UI
module ASM
  class Service
    def initialize(service, decrypt=true)
      @service = service
      @decrypt = decrypt
      @component_instances = nil
    end

    def to_s
      "#<ASM::Service name: %s id: %s>" % [deployment_name, id]
    end

    def teardown?
      !!@service["teardown"]
    end

    def migration?
      !!@service["migration"]
    end

    def retry?
      !!@service["retry"]
    end

    def raw_service
      @service
    end

    def deployment_name
      @service["deploymentName"]
    end

    def id
      @service["id"]
    end

    def template
      @service["serviceTemplate"]
    end

    # Generates a component programatically and inject it into the template
    #
    # @example generates a basic cluster resource
    #     service.generate_component("my_cluster.aidev.com") do |component|
    #       component.type = "CLUSTER"
    #       component.add_resource("asm::cluster", "Cluster Settings", [
    #         {:id => "datacenter", :value => "M830Datacenter"},
    #         {:id => "title", :value => "vcenter-env10-vcenter.aidev.com"}
    #       ])
    #     end
    #
    #     component = service.component_by_id("my_cluster.aidev.com")
    #
    # @see ASM::Service::Component::Generator
    # @param certname [String] a certificate name
    # @yieldparam generated [Component::Generator]
    # @return [void]
    # @raise [StandardError] any unhandled errors raised while calling Component::Generator
    def generate_component(certname)
      generated = Component::Generator.new(certname)
      yield(generated)

      component_hash = generated.to_component_hash

      template["components"] << component_hash

      if @component_instances
        @component_instances << Component.new(component_hash, @decrypt, self)
      end

      nil
    end

    def components
      @component_instances ||= template["components"].map do |c|
        Component.new(c, @decrypt, self)
      end
    end

    def each_component
      components.each do |component|
        yield(component)
      end
    end

    # Find a componant given its ID
    #
    # @param id [String] the component ID
    # @return [ASM::Service::Component]
    def component_by_id(id)
      components.select{|c| c.id == id}.first
    end

    # Finds components of a certain type
    #
    # @param type [String] a type like SERVER or CLUSTER
    # @return [Array<ASM::Service::Component>]
    def components_by_type(type)
      components.select{|c| c.type == type}
    end

    # Find related components in the service, optionally for a certain type
    #
    # It uses {ASM::Service::Component#related_components} to do the actual work
    # but this is a shortcut via the service rather than via component in a case
    # where a component might not know it's service
    #
    # This is a replacement for {ASM::ServiceDeployment#find_related_components}
    #
    # @param component [ASM::Service::Component] the companant to inspect
    # @param type [String] a type like SERVER or CLUSTER
    # @return [Array<ASM::Service::Component>]
    def related_components(component, type=nil)
      component.related_components(type, self)
    end
  end
end
