require 'asm/service'
require 'asm/util'
require 'asm/device_management'
require 'forwardable'

module ASM
  class Type
    class Base
      extend Forwardable

      # Within the providers for the type find a provider to handle a puppet type
      #
      # @raise [StandardError] when no provider can be found to handle the puppet type
      def self.provider_name(resource_type_name, logger)
        type = self.to_s.downcase.split("::").last.capitalize
        possible_providers = Type.providers.select{|p| p[:type] == type}
        possible_provider_classes = possible_providers.map{|p| p[:class]}

        logger.debug("About to look for a provider to handle %s within providers %s" % [resource_type_name, possible_provider_classes])

        possible_providers.each do |provider|
          if Array(provider[:class].puppet_type).include?(resource_type_name)
            if provider[:class].to_s =~ /^ASM::Provider::(.+)::(.+)$/
              logger.debug("Found provider %s for %s" % [provider[:class], resource_type_name])
              return($2)
            end
          end
        end

        logger.debug("Looked for a provider to handle %s within providers %s but could not find one" % [resource_type_name, possible_provider_classes])
        raise("Could not find a provider for resource type '%s' within types of class ASM::Type::%s" % [resource_type_name, type])
      end

      # Creates a Type instance from a Component
      #
      # @param component [ASM::Service::Component] the component to turn into a type
      # @return [ASM::Type::Base]
      def self.create(component, logger=ASM.logger)
        provider = nil
        resource = nil

        component.resource_ids.each do |resource_id|
          begin
            provider = provider_name(resource_id, logger)
            resource = component.configuration[resource_id]
          rescue
            next
          end

          break
        end

        raise("Could not find a provider to handle the configuration for component %s with resources %s using type %s" % [component.to_s, component.resource_ids.join(", "), self]) unless provider

        new(component, provider, resource, logger)
      end

      attr_reader :provider_config, :component, :service_component, :provider_name
      attr_accessor :deployment, :logger, :puppet_certname

      # @method process!
      # @method to_puppet
      def_delegators :provider, :process!, :to_puppet, :ensure, :ensure=, :uuid, :uuid=
      def_delegators :service_component, :component_id, :id, :teardown, :name, :guid, :type, :teardown?
      def_delegator  :service_component, :configuration, :component_configuration

      def initialize(component, provider_name, provider_config, logger=ASM.logger)
        @service_component = component
        @component = component.to_hash
        @provider_name = provider_name.downcase
        @provider_config = provider_config
        @logger = logger
        @puppet_certname = service_component.puppet_certname
      end

      def debug?
        deployment.debug?
      rescue
        ASM.config.debug_service_deployments
      end

      def service_teardown?
        service_component.service.teardown?
      end

      def type_name
        self.class.to_s.downcase.split("::").last
      end

      # @api private
      def get_provider_class(provider_name)
        klass_type = ::ASM::Provider.const_get(type_name.capitalize)
        klass_type.const_get(provider_name.capitalize)
      end

      def provider
        unless @provider
          klass = get_provider_class(@provider_name)

          @provider = klass.new
          @provider.type = self
          @provider.configure!(@provider_config)
        end

        @provider
      end

      # A helper to log and forward a method call to another object
      #
      # We often need to set up proxy methods in types where the type will call a method
      # on a provider.  It's good to log these calls for debug purposes to be sure we're
      # calling the right object and so forth but it's tedious to log every time.
      #
      # @example Typically we'd use a wrapper method like this:
      #    def remove_server_from_volume!(server)
      #       logger.debug("Calling remove_server_from_volume! for server %s" % server.to_s)
      #       provider.remove_server_from_volume!(server)
      #    end
      #
      # @example This will get tedious for types with many methods and yield inconsistent logging, so this helper lets you do:
      #    def remove_server_from_volume!(server)
      #      delegate(provider, :remove_server_from_volume!, server)
      #    end
      #
      # And will cause standard log lines to be produced that shows the caller, destination object
      # and method
      #
      # @api private
      # @raise Any exception from the called method
      # @return Any return value from the called method
      def delegate(target, method, *args)
        ("%s calling delegated method %s on %s" % [File.basename(caller[1]), method, target.to_s])
        logger.debug("%s calling delegated method %s on %s" % [File.basename(caller[1]), method, target.to_s])
        target.send(method, *args)
      end

      # Extracts a Service Tag from a standard format certificate name
      #
      # @return [String]
      def cert2serial(cert_name=nil)
        /^[^-]+-(.*)$/.match(cert_name || puppet_certname)[1].upcase
      end
      alias_method :cert2servicetag, :cert2serial

      # Determines if inventory should be run for a specific piece of hardware
      #
      # @return [Boolean]
      def should_inventory?
        provider.should_inventory?
      end

      # Updates the inventory of a device
      #
      # It will check using {#should_inventory?} if the device supports being inventoried
      #
      # This proxies to the provider, so different hardware can support different inventory methods
      #
      # @return (see ASM::Provider::Base#update_inventory)
      # @raise (see ASM::Provider::Base#update_inventory)
      def update_inventory
        if should_inventory?
          logger.info("Updating inventory for %s" % puppet_certname)
          @inventory = nil
          provider.update_inventory
        else
          logger.info("update_inventory called for %s but it does not support being inventoried" % puppet_certname)
        end
      end

      # Returned the cached inventory or retrieve and cache it
      #
      # use {#retrieve_inventory!} to force an update
      #
      # @return [Hash] in the format seen in spec/fixtures/asm_server_m620.json
      def retrieve_inventory
        @inventory || retrieve_inventory!
      end

      # Retrieves the inventory stored by ASM Server RA and update a local cache
      #
      # @return [Hash] in the format seen in spec/fixtures/asm_server_m620.json
      def retrieve_inventory!
        @inventory = Util.fetch_server_inventory(puppet_certname)
      end

      # Parses the device configuration
      #
      # @return [Hashie::Mash, nil] nil when the config cannot be found
      def device_config
        ASM::DeviceManagement.parse_device_config(puppet_certname)
      end

      # Run process_generic on the deployment and surpress running inventories
      #
      # Updating of inventories are to be done in a rule
      #
      # @param cert_name [String] the certificate name to process
      # @param config [Hash] a set of puppet resource asm process_node will accept
      # @param puppet_run_type ["device", "apply"] the way puppet should be run
      # @param override [Boolean] override config file settings
      # @param server_cert_name [String] allow customization of log file names for shared devices
      # @param asm_guid [String] internal ASM inventory GUID
      # @param update_inventory [Boolean] instruct the {ServiceDeployment#process_generic} to do inventory updates
      # @return [Hash] the results from the puppet run
      def process_generic(cert_name, config, puppet_run_type, override = true, server_cert_name = nil, asm_guid = nil, update_inventory=false)
        if !debug?
          result = deployment.process_generic(cert_name, config, puppet_run_type, override, server_cert_name, asm_guid, update_inventory)
        else
          require "pp"
          logger.info("Would have run process_generic for %s using %s" % [cert_name, puppet_run_type])
          logger.info("\n%s" % config.pretty_inspect)
          result = {}
        end

        result
      end

      def provider_path
        "%s/%s" % [type_name, @provider_name]
      end

      # Validates the relationships between this resource and another
      #
      # Given a resource it will extract the resource type from it and then run
      # a type specific check on the provider.  The current list of methods
      # being called are {ASM::Provider::Base#volume_supported?},
      # {ASM::Provider::Base#server_supported?}, {ASM::Provider::Base#cluster_supported?}
      # {ASM::Provider::Base#virtualmachine_supported?}
      #
      # This allow provider writers to declare their support for other related systems
      # for example ASM does not support using NetAPP volumes with HyperV so the
      # {ASM::Provider::Volume::Netapp#cluster_supported?} method enforces that
      # it only supports VMWare today
      #
      # @param resource [ASM::Type::Base] Any valid resource type
      # @return [Boolean]
      def supports_resource?(resource)
        logger.debug("Checking if provider %s supports a resource of provider %s" % [provider_path, resource.provider_path])

        case resource.type_name
        when "volume"
          provider.volume_supported?(resource)
        when "server"
          provider.server_supported?(resource)
        when "cluster"
          provider.cluster_supported?(resource)
        when "virtualmachine"
          provider.virtualmachine_supported?(resource)
        when "controller"
          provider.controller_supported?(resource)
        else
          raise("Do not know how to check if a resource of type %s is supported" % resource.type_name)
        end
      end

      def to_s
        "#<%s:%s:%s %s>" % [self.class, self.object_id, provider_path, puppet_certname]
      end

      # Find related components in the current deployment
      #
      # @param type [String] a type like CLUSTER, SERVER, STORAGE etc
      # @return [Array<ASM::Type::Base>] found components as resources
      def related_components(type)
        components = service_component.related_components(type)
        Type::to_resources(components, nil, @deployment, logger)
      end

      # Find the first related server
      #
      # @return [ASM::Type::Server, nil] nil when none are found
      def related_server
        related_servers.first
      end

      # Find the first related cluster
      #
      # @return [ASM::Type::Cluster, nil] nil when none are found
      def related_cluster
        related_clusters.first
      end

      # Find all related clusters
      #
      # @return [Array<ASM::Type::Cluster>]
      def related_clusters
        related_components("CLUSTER")
      end

      # Find all related servers
      #
      # @return [Array<ASM::Type::Server>]
      def related_servers
        related_components("SERVER")
      end

      # Find all related volumes
      #
      # @return [Array<ASM::Type::Volume>]
      def related_volumes
        related_components("STORAGE")
      end

      # Prepares a resource for teardown
      #
      # Steps that need to be taken to prepare a related resources for
      # teardown can go in here,  types can do more or less anything they
      # need in here including custom teardown steps.
      #
      # Should these steps remove the need for process! to be run or
      # just wish to short circuit that or prevent teardown they should
      # return false
      #
      # @return [Boolean] false to indicate process! should be skipped
      def prepare_for_teardown!
        delegate(provider, :prepare_for_teardown!)
      end
    end
  end
end
