require 'asm/type'
require 'asm/network_configuration'

module ASM
  class Type
    class Server < Base
      # Retrieves the asm::esxiscsiconfig resource for the server
      #
      # Just the parameters will be returned, {} when the server does not
      # have asm::esxiscsiconfig
      #
      # @return [Hash]
      def network_params
        if config = service_component.resource_by_id("asm::esxiscsiconfig")
          config.parameters
        else
          {}
        end
      end

      # Constructes a NetworkConfiguration object based on asm::esxiscsiconfig
      #
      # @return [ASM::NetworkConfiguration]
      def network_config
        raw_config = network_params["network_configuration"] || {"interfaces" => []}
        ASM::NetworkConfiguration.new(raw_config)
      end

      def management_network
        network_config.get_network("HYPERVISOR_MANAGEMENT")
      end

      def static_network_config
        management_network["staticNetworkConfiguration"]
      end

      def primary_dnsserver
        static_network_config["primaryDns"]
      end

      def secondary_dnsserver
        static_network_config["secondaryDns"]
      end

      def fcoe_san_networks
        network_config.get_networks("STORAGE_FCOE_SAN")
      end

      # Extract the serverType from the inventory
      #
      # @return ["RACK", "BLADE", "TOWER", "SLED", "COMPELLENT", nil]
      def physical_type
        provider.physical_type
      end

      # Determines if a server ia rackserver
      #
      # @return [Boolean]
      def rackserver?
        !!puppet_certname.match(/rackserver/)
      end

      # Determines if a server is a Fibre Channel over Ethernet server
      #
      # @return [Boolean]
      def fcoe?
        !fcoe_san_networks.empty?
      end

      # The hosts IP Address on the management network
      #
      # @todo rename for clarity
      def hostip
        static_network_config["ipAddress"]
      end
      alias_method :static_ipaddress, :hostip

      def hostname
        provider.hostname
      end

      def admin_password
        provider.admin_password
      end

      def os_image_type
        provider.os_image_type
      end

      def serial_number
        provider.serial_number
      end

      def serial_number=(serial)
        provider.serial_number = serial
      end

      # Determines if a server is configured to boot from SAN
      #
      # @return [Boolean]
      def boot_from_san?
        delegate(provider, :boot_from_san?)
      end

      # Determines if a server is being deployed with an OS
      #
      # @return [Boolean]
      def has_os?
        delegate(provider, :has_os?)
      end

      def razor_policy_name
        ("policy-%s-%s" % [hostname, deployment.id]).downcase
      end

      # Resets any Virtual MAC addresses back to permenent MAC addresses
      #
      # @return [void]
      def reset_mac_addresses!
        delegate(provider, :reset_mac_addresses!)
      end

      # Delete an associated puppet certificate for the host
      #
      # @return [void]
      # @raise [StandardError] when removal fails
      def delete_server_cert!
        delegate(provider, :delete_server_cert!)
      end

      # Delete the node data for server that contains the post-installation information
      #
      # @return [void]
      # @raise [StandardError] when removal fails
      def delete_server_node_data!
        delegate(provider, :delete_server_node_data!)
      end

      # Remove the server from any cluster it might belong to
      #
      # @return [void]
      # @raise [StandardError] when leaving the cluster fails
      def leave_cluster!
        delegate(provider, :leave_cluster!)
      end

      # Remove access rights from all volumes associated with this server
      #
      # @return [void]
      def clean_related_volumes!
        delegate(provider, :clean_related_volumes!)
      end

      # Clean up boot from SAN virtual identies
      #
      # A virtual identity is a virtual MAC and World Wide Port Name (WWPN)
      # which gets assigned to it from a ASM managed pool.  On termination the
      # node needs to be reset to factory values and the MAC and WWPN returned
      # to the pool
      #
      # @return [void]
      def clean_virtual_identities!
        delegate(provider, :clean_virtual_identities!)
      end

      # Powers off the particular machine
      #
      # @return [void]
      def power_off!
        delegate(provider, :power_off!)
      end

      # Extracts the HyperV configuration as a hash from a server
      #
      # Extracts a hash of all non nil value properties tagged as :hyperv in the provider.
      #
      # @raise [StandardError] on servers that are not HyperV compatible
      # @return [Hash] of non nil properties and values
      def hyperv_config
        provider.to_hash(true, :hyperv)
      end

      # Determine if a machine has been deployed based on the database and Razor
      #
      # @return [Boolean]
      def deployment_completed?
        begin
          return true if db_execution_status == "complete"
        rescue ASM::Data::NoExecution
        end

        return false if boot_from_san? || !has_os?

        ["boot_local", "boot_local_2"].include?(razor_status)
      end

      # Retrieves the component status from the database components table
      #
      # Uses {Data::Deployment#get_component_status}
      #
      # @raise [ASM::Data::NoExecution] when no execution is configured in the db
      # @return [String, nil] the component status from the database
      def db_execution_status
        begin
          status = deployment.db.get_component_status(id)
        rescue ASM::Data::NoExecution
          raise
        rescue
          logger.warn("Could not retrieve deployment status from the database for id %s: %s: %s" % [id, $!.class, $!.to_s])
          status = nil
        end

        status ? status[:status] : nil
      end

      # Retrieves the razor task status for the server
      #
      # Possible status values can be seen in {Razor#task_status}
      #
      # @return [Hash] the razor status returned from task_status
      def razor_status
        razor = ASM::Razor.new(:logger => logger)
        node = razor.find_node(cert2serial)

        status = {}

        if !node["name"].nil?
          status = razor.task_status(node["name"], razor_policy_name)
        end

        status
      rescue
        logger.debug("Razor policy for %s is not applied correctly or retrieval failed: %s: %s" % [puppet_certname, $!.class, $!.to_s])
        {}
      end

      # Extracts Fabric configuration from the server network configuation
      #
      # Based on the information found in the cards # found in {#network_config}
      # for rack servers it produce a Hash of # fabrics
      #
      # For rack servers the fabrics map 1:1 to the card slot index
      #
      # The number indicates a 2 port or 4 port card
      #
      # @example output
      #
      #     {"Fabric A"=>4, "Fabric B"=>2, "Fabric C"=>2}
      #
      # @return [Hash, nil] of Fabric information
      def fabric_info
        fabric_info = {}
        curr_fabric = "A"
        network_config.cards.each do |card|
          unless fabric = card.fabric_letter
            fabric = curr_fabric
            curr_fabric = curr_fabric.succ
          end
          fabric_info["Fabric %s" % fabric] = card.nictype.n_ports
        end

        if network_config.is_blade?
          # This is just to match previous behavior of get_server_nic_type; the values
          # here make no sense whatsoever, the user has not configured these fabrics
          ("A".."C").each do |fabric|
            fabric_info["Fabric %s" % fabric] ||= 2
          end
        end

        fabric_info
      end
    end
  end
end
