require 'spec_helper'
require 'asm/type'

ASM::Type.load_providers!

describe ASM::Provider::Cluster::Vmware do
  let(:logger) { stub(:debug => nil, :warn => nil, :info => nil) }
  let(:deployment) { stub(:id => "1234", :debug? => false, :decrypt? => true) }
  let(:service) { SpecHelper.service_from_fixture("Teardown_EQL_VMware_Cluster.json") }
  let(:server_components) { service.components_by_type("SERVER") }
  let(:volume_components) { service.components_by_type("STORAGE") }
  let(:cluster_component) { service.components_by_type("CLUSTER").first }

  let(:type) { cluster_component.to_resource(deployment, logger) }
  let(:provider) { type.provider }

  let(:server) { server_components.first.to_resource(deployment, logger) }

  before(:each) do
    deployment.stubs(:lookup_hostname).with(server.hostname, server.static_network_config).returns("rspec-testhost")

    type.stubs(:related_servers).returns([server])
    provider.stubs(:host_username).returns("root")
  end

  describe "#should_inventory?" do
    it "should only do inventories when there is a guid" do
      expect(provider.should_inventory?).to be(true)

      type.expects(:guid).returns(nil)
      expect(provider.should_inventory?).to be(false)
    end
  end

  describe "#update_inventory" do
    it "should update using update_asm_inventory" do
      ASM::Util.expects(:update_asm_inventory).with(cluster_component.guid)
      provider.update_inventory
    end

    it "should fail without a guid" do
      type.expects(:guid).returns(nil)
      expect { provider.update_inventory }.to raise_error("Cannot update inventory for vcenter-env10-vcenter.aidev.com without a guid")
    end
  end

  describe "#virtualmachine_supported?" do
    it "should support vmware virtual machines" do
      expect(provider.virtualmachine_supported?(stub(:provider_path => "virtualmachine/vmware"))).to be(true)
    end

    it "should not support other virtual machines" do
      expect(provider.virtualmachine_supported?(stub(:provider_path => "rspec/rspec"))).to be(false)
    end
  end

  describe "#asm_host_hash" do
    it "should create just a asm::host hash by default without a transport" do
      hash = provider.asm_host_hash("cert", "host", "user", "pass", "present")

      expected_properties = {
        "datacenter"=>"M830Datacenter",
        "cluster"=>"M830Cluster",
        "hostname"=> "host",
        "username"=>"user",
        "password"=>"pass",
        "decrypt"=>true,
        "timeout"=>90,
        "ensure" => "present"
      }

      expect(hash).to include("asm::host")
      expect(hash).to_not include("transport")
      expect(hash["asm::host"]["cert"]).to eq(expected_properties)
    end

    it "should be able to include the transport" do
      hash = provider.asm_host_hash("cert", "host", "user", "pass", "present", true)
      expect(hash).to include("transport")
    end
  end

  describe "#transport_config" do
    it "should create a valid transport config" do
      hash = provider.transport_config

      expected = {
        "transport"=> {
          "vcenter"=>{
            "name"=>"vcenter-env10-vcenter.aidev.com",
            "options"=>{"insecure"=>true},
            "provider"=>"device_file"
          }
        }
      }

      expect(hash).to eq(expected)
    end
  end

  describe "#evict_volume!" do
    it "should not remove servers being torn down" do
      servers = [mock(:teardown? => true), mock(:teardown? => true)]
      type.expects(:related_servers).returns(servers)
      type.expects(:process_generic).never
      provider.evict_volume!(stub)
    end

    it "should remove the volume from every server not being torn down" do
      volume = volume_components.first.to_resource(deployment, logger)
      server.expects(:teardown?).returns(false)
      volume.expects(:esx_datastore).returns({"esx_datastore" => "rspec"})

      expected = {"esx_datastore" => "rspec",
                  "transport" =>
                    {"vcenter" =>
                      {"name" => "vcenter-env10-vcenter.aidev.com", "options" =>
                        {"insecure" => true},"provider" => "device_file"
                      }
                    }
      }

      type.expects(:process_generic).with("vcenter-env10-vcenter.aidev.com", expected, "apply", true, nil, "ff8080814dbf2d1d014dc2a280fd011f")

      provider.evict_volume!(volume)
    end
  end

  describe "#evict_server!" do
    it "should evict the host from the cluster via process_generic" do
      server_hash = {'asm::host' => {},
        'transport' => {
          'vcenter' => {
          'name' => "vcenter-env10-vcenter.aidev.com",
          'options' => {'insecure' => true},
          'provider' => 'device_file'}}
      }

      server_hash['asm::host'][server.puppet_certname] = {
        'datacenter' => 'M830Datacenter',
        'cluster' => 'M830Cluster',
        'hostname' => "rspec-testhost",
        'username' => 'root',
        'password' => 'ff8080814dbf2d1d014ddcc519673595',
        'decrypt' => true,
        'timeout' => 90,
        'ensure' => 'absent'
      }

      type.expects(:process_generic).with(server.puppet_certname, server_hash, "apply", true, nil, nil)

      type.evict_server!(server)
    end
  end
end
