require 'spec_helper'
require 'asm/type'

ASM::Type.load_providers!

describe ASM::Provider::Server::Server do
  let(:service) { SpecHelper.service_from_fixture("Teardown_EQL_VMware_Cluster.json") }
  let(:logger) { stub(:debug => nil, :warn => nil, :info => nil) }
  let(:deployment) { stub(:id => "1234", :is_teardown? => true, :debug? => false, :log => nil) }
  let(:volume_components) { service.components_by_type("STORAGE") }
  let(:server_components) { service.components_by_type("SERVER") }
  let(:cluster_components) { service.components_by_type("CLUSTER") }
  let(:server_component) { server_components.first }

  let(:type) { server_component.to_resource(deployment, logger) }
  let(:provider) { type.provider }

  let(:server_hostname) { server_component.configuration["asm::server"].to_a[0][1]["os_host_name"] }
  let(:server_puppet_certname) { server_component.configuration["asm::server"].to_a[0][0] }

  describe "#should_inventory?" do
    it "should only do inventories when there is a guid" do
      expect(provider.should_inventory?).to be(true)

      type.expects(:guid).returns(nil)
      expect(provider.should_inventory?).to be(false)
    end
  end

  describe "#update_inventory" do
    it "should update using update_asm_inventory" do
      ASM::Util.expects(:update_asm_inventory).with(server_component.guid)
      provider.update_inventory
    end

    it "should fail without a guid" do
      type.expects(:guid).returns(nil)
      expect { provider.update_inventory }.to raise_error("Cannot update inventory for bladeserver-15kvd42 without a guid")
    end
  end

  describe "#network_params" do
    it "should return the params when present" do
      expect(type.network_params).to eq(server_component.resource_by_id("asm::esxiscsiconfig").parameters)
    end

    it "should return empty otherwise" do
      server_component.expects(:resource_by_id).with("asm::esxiscsiconfig").returns(nil)
      expect(type.network_params).to eq({})
    end
  end

  describe "when tearing down" do
    describe "#to_puppet" do
      it "should prepare the correct teardown hash" do
        teardown_hash = {
            'asm::server' => {server_puppet_certname => {
              'admin_password' => 'ff8080814dbf2d1d014ddcc519673595',
              'broker_type' => 'noop',
              'decrypt' => true,
              'ensure' => 'absent',
              'esx_mem' => 'false',
              'installer_options' => {
                  "ntp_server" => "172.20.0.8",
                  "os_type"=>"vmware_esxi",
                  "agent_certname"=>"agent-esx2ip1"
              },
              'os_host_name' => server_hostname,
              'os_image_type' => 'vmware_esxi',
              'policy_name' => 'policy-esx2ip1-1234',
              'serial_number' => '15KVD42',
              'razor_image' => 'esxi-5.5',
              'razor_api_options' => {'url' => 'http://asm-razor-api:8080/api'}}
            }
        }

        expect(provider.to_puppet["asm::server"]).to eq(teardown_hash["asm::server"])
      end
    end
  end

  describe "#create_idrac_resource" do
    it "should create the correct idrac controller resource" do
      idrac = provider.create_idrac_resource

      expect(idrac).to be_a(ASM::Type::Controller)
      expect(idrac.provider).to be_a(ASM::Provider::Controller::Idrac)
      expect(idrac.puppet_certname).to eq(type.puppet_certname)
    end
  end

  describe "#cluster_supported?" do
    it "should check hyperv is configured correctly do" do
      provider.expects(:configured_for_hyperv?).returns(false)
      expect(provider.configured_for_hyperv?).to be(false)

      provider.expects(:configured_for_hyperv?).returns(true)
      expect(provider.configured_for_hyperv?).to be(true)
    end

    it "should support other clusters by default" do
      expect(provider.cluster_supported?(stub(:provider_path => "rspec"))).to be(true)
    end
  end

  describe "#configured_for_hyperv?" do
    it "should correctly detect missing properties" do
      needed = ASM::Provider::Cluster::Scvmm::SERVER_REQUIRED_PROPERTIES

      provider.stubs(:to_hash).with(true, :hyperv).returns({needed.first => "foo"})
      logger.expects(:warn).with("Server bladeserver-15kvd42 is not supported by HyperV as it lacks these properties or they have nil values: %s" % [needed[1..-1].join(", ")])

      expect(provider.configured_for_hyperv?).to eq(false)
    end

    it "should detect nil values" do
      needed = ASM::Provider::Cluster::Scvmm::SERVER_REQUIRED_PROPERTIES
      has = Hash[needed[1..-1].map {|k| [k, "rspec"]}]

      provider.stubs(:to_hash).with(true, :hyperv).returns(has)
      logger.expects(:warn).with("Server bladeserver-15kvd42 is not supported by HyperV as it lacks these properties or they have nil values: %s" % [needed.first])

      expect(provider.configured_for_hyperv?).to eq(false)
    end

    it "should report ok when there are no missing values" do
      needed = ASM::Provider::Cluster::Scvmm::SERVER_REQUIRED_PROPERTIES
      has = Hash[needed.map {|k| [k, "rspec"]}]

      provider.stubs(:to_hash).with(true, :hyperv).returns(has)
      logger.expects(:warn).never

      expect(provider.configured_for_hyperv?).to eq(true)
    end
  end

  describe "#configure_installer_options" do
    it "should not configure agent_certname if no hostname is set" do
      provider.installer_options = {}

      provider.expects(:hostname).returns(nil)
      provider.configure_installer_options
      expect(provider.installer_options).to_not include("agent_certname")
    end

    it "should construct installer options based on the additional properties" do
      provider.installer_options = {}
      provider.configure_installer_options

      expect(provider.installer_options).to eq({
        "ntp_server" => "172.20.0.8",
        "os_type"=>"vmware_esxi",
        "agent_certname"=>"agent-esx2ip1"
      })
    end
  end

  describe "#configure_hook" do
    it "should configure the installer options" do
      provider.expects(:configure_installer_options).once
      provider.configure_hook
    end
  end

  describe "#to_puppet" do
    it "should remove resources that are not puppet types" do
      config = type.component_configuration
      config["asm::bios"] = {}
      config["asm::baseserver"] = {}
      type.expects(:component_configuration).returns(config)

      expect(provider.to_puppet.keys).to eq(["asm::server"])
    end
  end

  describe "#hostname" do
    it "should return the os_host_name property as hostname" do
      expect(provider.hostname).to eq(server_hostname)
    end
  end

  describe "#hostname_to_certname" do
    it "should should scrub illegal characters from the hostname" do
      expect(provider.hostname_to_certname("*'_\/!~`@#%^&()$rspec")).to eq("agent-rspec")
    end

    it "should support taking the resource hostname by default" do
      expect(provider.hostname_to_certname).to eq("agent-esx2ip1")
    end
  end

  describe "#dell_server?" do
    it "should check the server certname by default" do
      ASM::Util.expects(:dell_cert?).with(server_puppet_certname).once
      provider.dell_server?
    end

    it "should support checking a supplied certname" do
      ASM::Util.expects(:dell_cert?).with("rspec").once
      provider.dell_server?("rspec")
    end
  end

  describe "#target_boot_device" do
    it "Should be nil for servers without asm::idrac" do
      server_component.expects(:resource_by_id).with("asm::idrac").returns(nil)
      expect(provider.target_boot_device).to be(nil)
    end

    it "should be nil for servers with asm::idrac but not target_boot_device" do
      server_component.expects(:resource_by_id).with("asm::idrac").returns({})
      expect(provider.target_boot_device).to be(nil)
    end
  end

  describe "#boot_from_san?" do
    it "should be false for non dell servers" do
      provider.expects(:dell_server?).returns(false)
      provider.expects(:target_boot_device).never
      expect(provider.boot_from_san?).to be(false)
    end

    it "should be true if target_boot_device is a boot from san one" do
      provider.expects(:target_boot_device).returns(ASM::Provider::Server::Server::BOOT_FROM_SAN_TARGETS[0])
      expect(provider.boot_from_san?).to be(true)
    end

    it "should be false if target_boot_device is not a boot from san one" do
      expect(provider.boot_from_san?).to be(false)
    end
  end

  describe "#idrac?" do
    it "should correctly report a server as being idrac when asm::idrac is present" do
      expect(provider.idrac?).to eq(true)
    end

    it "should correctly report a server as not being idrac when asm::idrac is absent" do
      type.service_component.expects(:has_resource_id?).with("asm::idrac").returns(false)
      expect(provider.idrac?).to eq(false)
    end
  end

  describe "#delete_server_cert!" do
    it "should not clean servers without an os_image_type" do
      ASM::DeviceManagement.expects(:clean_cert).never

      provider.os_image_type = nil
      provider.delete_server_cert!

      provider.os_image_type = ""
      provider.delete_server_cert!
    end

    it "should clean certs via DeviceManagement" do
      ASM::DeviceManagement.expects(:clean_cert).with(provider.hostname_to_certname(server_hostname))
      provider.delete_server_cert!
    end
  end

  describe "#delete_server_node_data!" do
    it "should not clean servers without an os_image_type" do
      ASM::DeviceManagement.expects(:remove_node_data).never

      provider.os_image_type = nil
      provider.delete_server_node_data!

      provider.os_image_type = ''
      provider.delete_server_node_data!
    end

    it "should clean certs via DeviceManagement" do
      ASM::DeviceManagement.expects(:remove_node_data).with(provider.hostname_to_certname(server_hostname))
      provider.delete_server_node_data!
    end
  end

  describe "#leave_cluster!" do
    it "should request the cluster remove the server when an association is found" do
      cluster = cluster_components.first.to_resource(deployment, logger)
      type.expects(:related_cluster).returns(cluster)

      cluster.expects(:evict_server!).with(type)
      cluster.expects(:teardown?).returns(false)

      type.leave_cluster!
    end

    it "should not remove itself from a cluster that is also being removed" do
      cluster = cluster_components.first.to_resource(deployment, logger)
      type.expects(:related_cluster).returns(cluster)

      cluster.expects(:evict_server!).never
      cluster.expects(:teardown?).returns(true)

      type.leave_cluster!
    end

    it "should do nothing otherwise" do
      type.expects(:related_cluster).returns(nil)
      type.leave_cluster!
    end
  end

  describe "#clean_related_volumes!" do
    it "should do nothing when the server is not an idrac one" do
      provider.expects(:idrac?).returns(false)
      type.expects(:related_volumes).never
      provider.clean_related_volumes!
    end

    it "should remove the server associations from any related volumes that are not being torn down" do
      volumes = [
        mock(:teardown? => false, :puppet_certname => "rspec_volume"),
        mock(:teardown? => true)
      ]

      volumes[0].expects(:remove_server_from_volume!).with(type).once
      type.expects(:related_volumes).returns(volumes)

      provider.expects(:idrac?).returns(true)

      provider.clean_related_volumes!
    end

    it "should squash failures and continue trying other volumes" do
      volumes = [
        stub(:teardown? => false, :puppet_certname => "rspec_volume0"),
        stub(:teardown? => false, :puppet_certname => "rspec_volume1")
      ]

      volumes[0].expects(:remove_server_from_volume!).raises("rspec simulation")
      volumes[1].expects(:remove_server_from_volume!).returns(nil)

      type.expects(:related_volumes).returns(volumes)
      provider.expects(:idrac?).returns(true)
      logger.expects(:warn).with("Failed to remove the server bladeserver-15kvd42 from the volume rspec_volume0: RuntimeError: rspec simulation")
      logger.expects(:warn).with(regexp_matches(/Failed to remove .+ rspec_volume1/)).never

      provider.clean_related_volumes!
    end
  end

  describe "#clean_virtual_identities!" do
    it "should do nothing when the server is not an idrac one" do
      provider.expects(:create_idrac_resource).returns(false)
      provider.clean_virtual_identities!
    end

    it "should do the cleanup for idrac servers" do
      provider.expects(:wait_for_lc_ready)

      idrac = provider.create_idrac_resource
      idrac.expects(:ensure=).with("teardown")
      idrac.expects(:process!)
      provider.expects(:create_idrac_resource).returns(idrac)

      provider.clean_virtual_identities!
    end
  end

  describe "#power_off!" do
    it "should use wsman to power off a dell server" do
      type.expects(:device_config).returns(device_config = mock)
      ASM::WsMan.expects(:poweroff).with(device_config, logger)
      provider.expects(:dell_server?).returns(true)
      provider.expects(:wait_for_lc_ready)

      provider.power_off!
    end

    it "should use IPMI to power off non dell servers" do
      type.expects(:device_config).returns(device_config = mock)
      ASM::Ipmi.expects(:power_off).with(device_config, logger)
      provider.expects(:dell_server?).returns(false)

      provider.power_off!
    end
  end
end

