require 'spec_helper'
require 'asm/type'

ASM::Type.load_providers!

describe ASM::Provider::Virtualmachine::Vmware do
  let(:logger) { stub(:debug => nil, :warn => nil, :info => nil) }
  let(:service) { SpecHelper.service_from_fixture("Teardown_CMPL_VMware_Cluster.json") }
  let(:deployment) { stub(:id => "1234", :is_teardown? => true, :debug? => false, :log => nil) }
  let(:vm_component) { service.components_by_type("VIRTUALMACHINE")[0] }
  let(:cluster_components) { service.components_by_type("CLUSTER") }

  let(:type) { vm_component.to_resource(deployment, logger) }
  let(:provider) { type.provider }

  describe "#prepare_for_teardown!" do
    it "should attempt to clean the server certificate and swallow errors" do
      provider.server.expects(:delete_server_cert!).raises("rspec")
      provider.prepare_for_teardown!
    end
  end

  describe "#cluster_supported?" do
    it "should support vmare clusters" do
      expect(provider.cluster_supported?(cluster_components.first.to_resource(deployment, logger))).to be(true)
    end

    it "should not support other clusters do" do
      expect(provider.cluster_supported?(stub(:provider_path => "rspec/rspec"))).to be(false)
    end
  end

  describe "#hostname" do
    it "should take the configured hostname if it has one" do
      provider.hostname = "rspec"
      expect(provider.hostname).to eq("rspec")
    end

    it "should allow overriding by the associated server" do
      provider.hostname = nil
      provider.server.stubs(:hostname).returns("rspec")
      expect(provider.hostname).to eq("rspec")
    end

    it "should use the hostname otherwise" do
      provider.server = nil
      provider.hostname = "rspec"
      expect(provider.hostname).to eq("rspec")
    end
  end

  describe "#additional_resources" do
    it "should include the modified asm::server when a server exist" do
      provider.server.stubs(:to_puppet).returns({"asm::server" => {"rspec" => {}}})
      expect(provider.additional_resources).to eq({"asm::server" => {"rspec" => {}}})
    end

    it "should add no additional resources otherwise" do
      provider.server = nil
      expect(provider.additional_resources).to eq({})
    end
  end

  describe "#configure_guest_type!" do
    before(:each) do
      provider.os_type = nil
      provider.os_guest_id = nil
      provider.scsi_controller_type = "BusLogic Parallel"
    end

    it "should set the correct windows options" do
      provider.server.stubs(:os_image_type).returns("rspec_windows_rspec")
      provider.configure_guest_type!

      expect(provider.os_type).to eq("windows")
      expect(provider.os_guest_id).to eq("windows8Server64Guest")
      expect(provider.scsi_controller_type).to eq("LSI Logic SAS")
    end

    it "should set the correct linux options" do
      provider.server.stubs(:os_image_type).returns("rspec_linux_rspec")
      provider.configure_guest_type!

      expect(provider.os_type).to eq("linux")
      expect(provider.os_guest_id).to eq("rhel6_64Guest")
      expect(provider.scsi_controller_type).to eq("VMware Paravirtual")
    end
  end

  describe "#configure_related_cluster!" do
    it "should configure the cluster properties from the related cluster" do
      provider.cluster = nil
      provider.datacenter = nil
      provider.vcenter_id = nil
      provider.vcenter_options = {}

      provider.configure_related_cluster!

      cluster = type.related_cluster

      expect(provider.cluster).to eq(cluster.provider.cluster)
      expect(provider.datacenter).to eq(cluster.provider.datacenter)
      expect(provider.vcenter_id).to eq(cluster.puppet_certname)
      expect(provider.vcenter_options).to eq({"insecure" => true})
    end
  end

  describe "#configure_network!" do
    it "should clear the network_interfaces on teardown" do
      provider.network_interfaces = []
      provider.configure_network!
      expect(provider.network_interfaces).to eq(nil)
    end

    it "should warn otherwise" do
      provider.ensure = "present"
      type.stubs(:teardown?).returns(false)
      logger.expects(:warn).with("Cannot configure networking for Virtual Machine vm-linuxvm1 as it only supports teardown")
      provider.configure_network!
    end
  end

  describe "#create_server_resource" do
    it "should create a server resource" do
      expect(provider.create_server_resource).to be_a(ASM::Type::Server)
    end

    it "should return nil when it fails" do
      ASM::Service::Component::Resource.any_instance.stubs(:to_component).raises("rspec simulation")
      expect(provider.create_server_resource).to eq(nil)
    end
  end

  describe "#configure_hook" do
    before(:each) do
      provider.stubs(:configure_server!).returns(true)
      provider.stubs(:configure_guest_type!)
      provider.stubs(:configure_related_cluster!)
      provider.stubs(:configure_network!)
    end

    it "should configure the guest type when a server was found" do
      provider.expects(:configure_server!).returns(true)
      provider.expects(:configure_guest_type!)
      provider.configure_hook
    end

    it "should not configure the guard type when a server was not found" do
      provider.expects(:configure_server!).returns(false)
      provider.expects(:configure_guest_type!).never
      provider.configure_hook
    end
  end

  describe "#configure_server!" do
    it "should configure the types when a server is found" do
      provider.server.serial_number = "rspec"
      provider.server.stubs(:hostname).returns("rspec")
      provider.hostname = nil

      provider.configure_server!

      expect(provider.server.serial_number).to eq("")
      expect(provider.hostname).to eq("rspec")
      expect(provider.uuid).to eq("rspec")
      expect(provider.server.provider.uuid).to eq("rspec")
    end

    it "should do nothing otherwise" do
      provider.server = nil
      provider.expects(:create_server_resource).returns(false)
      provider.expects(:uuid=).never
      expect(provider.configure_server!).to eq(false)
    end
  end
end
