require 'spec_helper'
require 'asm/type'

ASM::Type.load_providers!

describe ASM::Provider::Volume::Compellent do
  let(:service) { SpecHelper.service_from_fixture("Teardown_CMPL_VMware_Cluster.json") }
  let(:logger) { stub(:debug => nil, :warn => nil, :info => nil) }
  let(:deployment) { stub(:id => "1234", :is_teardown? => true, :debug? => false, :log => nil) }
  let(:volume_component) { service.components_by_type("STORAGE")[0] }
  let(:server_components) { service.components_by_type("SERVER") }
  let(:cluster_components) { service.components_by_type("CLUSTER") }

  let(:type) { volume_component.to_resource(deployment, logger) }
  let(:provider) { type.provider }

  let(:expected_esx_datastore) do
    {"esx_datastore" =>
      {"172.28.10.60:ESXIFCCluster1" =>
        {"ensure"=>"absent",
         "datastore"=> "ESXIFCCluster1",
         "type"=>"vmfs",
         "lun"=>"RSPEC_LUNID",
         "transport"=>"Transport[vcenter]"
        }
      }
    }
  end

  describe "#initialize" do
    it "should specify the device run type" do
      expect(provider.puppet_type).to eq("asm::volume::compellent")
    end
  end

  describe "#clean_access_rights!" do
    it "should create compellent_server objects for any server being torn down" do
      type.stubs(:related_servers).returns([
        mock(:cert2serial => "server1", :puppet_certname => "server1", :teardown? => true),
        mock(:cert2serial => "server2", :puppet_certname => "server2", :teardown? => true)
      ])

      expected = {'compellent_server' =>
                  {'ASM_server1' =>
                   {'ensure' => 'absent',
                    'serverfolder' => ''},
                   'ASM_server2' =>
                   {'ensure' => 'absent',
                    'serverfolder' => ''}
                  }
      }

      type.expects(:process_generic).with("compellent-24260", expected, "apply", true, nil, "compellent-172.17.10.40")

      provider.clean_access_rights!
    end

    it "should not run puppet when there are no servers" do
      type.stubs(:related_servers).returns([])
      type.expects(:process_generic).never
      provider.clean_access_rights!
    end
  end

  describe "#compellent_server_resource" do
    it "should construct the right hash based on the server" do
      server = server_components.first.to_resource(deployment, logger)

      expected_data = {"compellent_server" => {"ASM_DP181Y1" => {"ensure" => "absent", "serverfolder" => ""}}}

      expect(provider.compellent_server_resource(server)).to eq(expected_data)
    end
  end

  describe "#remove_server_from_volume!" do
    it "should process_generic with the right data" do
      server = server_components.first.to_resource(deployment, logger)

      expected_data = {"compellent_server" => {"ASM_DP181Y1" => {"ensure" => "absent", "serverfolder" => ""}}}
      type.expects(:process_generic).with(type.puppet_certname, expected_data, "apply", true, nil, type.guid)

      provider.remove_server_from_volume!(server)
    end

    it "should raise failures from process_generic" do
      type.expects(:process_generic).raises("rspec simulation")
      server = server_components.first.to_resource(deployment, logger)
      expect { provider.remove_server_from_volume!(server) }.to raise_error("rspec simulation")
    end
  end

  describe "#esx_datastore_hash" do
    before :each do
      ASM::Util.stubs(:find_compellent_volume_info).returns("RSPEC_DEVICE_ID")
      ASM::Cipher.stubs(:decrypt_string).returns("DECRYPTED_STRING")
      deployment.stubs(:get_compellent_lunid).with(type.related_server.hostip, 'root', "DECRYPTED_STRING", "RSPEC_DEVICE_ID").returns("RSPEC_LUNID")
    end

    it "should create a valid datastore config" do
      expect(provider.esx_datastore(type.related_server, type.related_cluster, "absent")).to eq(expected_esx_datastore)
    end

    describe "when leaving a cluster" do
      it "should delegate to the cluster" do
        cluster = type.related_cluster
        cluster.stubs(:teardown?).returns(false)
        type.stubs(:related_cluster).returns(cluster)

        expected_hash = {}.merge(expected_esx_datastore).merge(cluster.provider.transport_config)

        servers = type.related_servers
        cluster.stubs(:related_servers).returns(servers)

        servers.each do |server|
          server.stubs(:teardown?).returns(false)
       end

        cluster.expects(:process_generic).with("vcenter-env10-vcenter.aidev.com", expected_hash, "apply", true, nil, "ff8080814dbf2d1d014dc2a280fd011f")
        type.leave_cluster!
      end
    end
  end
end
