require 'asm/service_deployment'
require 'json'
require 'spec_helper'
require 'yaml'
require 'asm/util'

describe ASM::ServiceDeployment do
  let(:puppetdb) { stub(:successful_report_after? => true) }

  before do
    ASM.init_for_tests
    @tmp_dir = Dir.mktmpdir
    @deployment_db = mock('deploymentdb')
    @deployment_db.stub_everything
    @sd = ASM::ServiceDeployment.new('8000', @deployment_db)
    razor = mock('razor')
    razor.stubs(:find_node).returns({})
    razor.stubs(:block_until_task_complete).returns({:status=>:boot_install, :timestamp=>Time.new("1969-01-01 00:00:00 -0600")})
    @sd.stubs(:razor).returns(razor)
    @sd.stubs(:puppetdb).returns(puppetdb)
    @sd.stubs(:create_broker_if_needed).returns('STUB-BROKER-NAME')
    @sd.stubs(:get_server_inventory).returns({})
    @sd.stubs(:update_inventory_through_controller)
    @sd.stubs(:reboot_all_servers).returns(nil)
    ASM.stubs(:base_dir).returns(@tmp_dir)
    network = {
      'id' => '1', 'name' => 'Test Network', 'vlanId' => '28',
      'staticNetworkConfiguration' => {
        'gateway' => '172.28.0.1', 'subnet' => '255.255.0.0'
      }
    }
    ASM::Util.stubs(:fetch_managed_inventory).returns([])
    mock_command_result = Hashie::Mash.new({
      'stdout' => '', 'stderr' => '', 'exit_status' => 0, 'pid' => 0
    })
    ASM::Util.stubs(:write_node_data)
    ASM::Util.stubs(:run_command_simple).returns(mock_command_result)
    ASM::Util.stubs(:run_command_success).returns(mock_command_result)
  end

  after do
    ASM.reset
  end

  describe 'when data is valid' do
    before do
      FileUtils.mkdir_p("#{@tmp_dir}/8000/resources")
      @r_dir = "#{@tmp_dir}/8000/resources"
      @r_file = "#{@r_dir}/server-cert.yaml"
      @o_file = "#{@tmp_dir}/8000/server-cert.out"
      @summary_file = "#{@r_dir}/state/server-cert/last_run_summary.yaml"
      @data = {'serviceTemplate' => {'components' => [
        {'id' => 'id', 'puppetCertName' => 'server-cert', 'resources' => []}
      ]}}
    end

    it 'should fail if the puppet run failed but exited with 0' do
      File.open( "#{@tmp_dir}/8000/server-cert.out", 'w') do |fh|
        fh.write('Results: For 0 resources. 0 from our run failed. 0 not from our run failed. 0 updated successfully.')
      end
      @sd.stubs(:iterate_file).with(@o_file).returns(@o_file)
      @sd.stubs(:iterate_file).with(@r_file).returns(@r_file)
      ASM::Util.expects(:run_command_streaming).with(
        "sudo puppet asm process_node --debug --trace --filename #{@r_file} --run_type apply --statedir #{@tmp_dir}/8000/resources --always-override server-cert", "#{@o_file}")
      @data['serviceTemplate']['components'][0]['type'] = 'TEST'
      @data['serviceTemplate']['components'][0]['resources'].push(
        {'id' => 'user', 'parameters' => [
          {'id' => 'title', 'value' => 'foo'},
          {'id' => 'foo', 'value' => 'bar'}
        ]}
      )

      ASM::DeviceManagement.expects(:puppet_run_success?).with("server-cert", 0, instance_of(Time), @o_file, @summary_file).returns([false, "rspec simulated fail"])

      expect { @sd.process(@data) }.to raise_error("puppet asm process_node for server-cert failed: rspec simulated fail")
    end

    it 'should be able to process data for a single resource' do
      File.open( "#{@tmp_dir}/8000/server-cert.out", 'w') do |fh|
        fh.write('Results: For 0 resources. 0 from our run failed. 0 not from our run failed. 0 updated successfully.')
      end
      @sd.stubs(:iterate_file).with(@o_file).returns(@o_file)
      @sd.stubs(:iterate_file).with(@r_file).returns(@r_file)
      ASM::DeviceManagement.stubs(:parse_device_config).with('server-cert').returns( {:host => '172.17.5.170' })
      ASM::Util.expects(:run_command_streaming).with(
        "sudo puppet asm process_node --debug --trace --filename #{@r_file} --run_type apply --statedir #{@tmp_dir}/8000/resources --always-override server-cert", "#{@o_file}")
      @data['serviceTemplate']['components'][0]['type'] = 'TEST'
      @data['serviceTemplate']['components'][0]['resources'].push(
        {'id' => 'user', 'parameters' => [
          {'id' => 'title', 'value' => 'foo'},
          {'id' => 'foo', 'value' => 'bar'}
        ]}
      )

      ASM::DeviceManagement.expects(:puppet_run_success?).with("server-cert", 0, instance_of(Time), @o_file, @summary_file).returns([true, "x"])

      @sd.process(@data)
      YAML.load_file(@r_file)['user']['foo']['foo'].should == 'bar'
    end

    describe 'for server bare metal provisioning' do
      it 'should configure a server' do

        @sd.stubs(:iterate_file).with(@o_file).returns(@o_file)
        @sd.stubs(:iterate_file).with(@r_file).returns(@r_file)

        ASM::Util.expects(:run_command_streaming).at_most(2).with(
            "sudo -i puppet asm process_node --filename #{@r_file} --run_type apply --always-override server-cert", "#{@o_file}") do |cmd|
          File.open(@o_file, 'w') do |fh|
            fh.write('Results: For 0 resources. 0 from our run failed. 0 not from our run failed. 0 updated successfully.')
          end
        end


        @data['serviceTemplate']['components'][0]['type'] = 'SERVER'
        @data['serviceTemplate']['components'][0]['resources'].push(
            {'id' => 'asm::server', 'parameters' => [
                {'id' => 'title', 'value' => 'server-cert'},
                {'id' => 'admin_password', 'value' => 'foo'},
                {'id' => 'os_host_name', 'value' => 'foo'},
                {'id' => 'os_image_type', 'value' => 'foo'}
            ]}
        )
        @sd.stubs(:is_fcoe_enabled).returns(false)
        @sd.stubs(:server_already_deployed).returns(false)
        ASM::DeviceManagement.stubs(:parse_device_config).with('server-cert').returns( {:host => '172.17.5.170' })
        ASM::DeviceManagement.expects(:puppet_run_success?).at_most(2).with("server-cert", 0, instance_of(Time), @o_file, @summary_file).returns([true, "x"])
        @sd.process(@data)
      end

      describe 'hyperV server' do
        it 'should process hyperv servers' do
          component =  {'id' => 'id', 'resources' => []}
          node = {'policy' => { 'name' => 'policy_test' } }
          @sd.stubs(:find_node).returns(node)
          policy = {
              'repo' => {'name' => 'esxi-5.1'},
              'installer' => {'name' => 'vmware_esxi'}
          }
          @sd.stubs(:get).returns(policy)
          component['id'] = 'id'
          component['puppetCertName'] = 'bladeserver-serialno'
          component['type'] = 'SERVER'
          parameters = [ {'id' => 'title', 'value' => 'bladeserver-serialno'},
                         {'id' => 'os_image_type', 'value' => 'hyperv'},
                         {'id' => 'os_image_version', 'value' => 'hyperv'},
                         {'id' => 'os_host_name', 'value' => 'foo'}
          ]
          resource1 = { 'id' => 'asm::server', 'parameters' => parameters }
          @sd.debug = true
          component['resources'].push(resource1)

          component['relatedComponents'] = { 'k1' => 'v1' }
          @sd.stubs(:server_already_deployed).returns(false)
          @sd.set_components_by_type('STORAGE',
                                     [
                                         {'id' => 'k1',
                                          'puppetCertName' => 'k1',
                                          'asmGUID' => 'equallogic-1.1.1.1',
                                          'componentID'=>'s1',
                                          'resources' => [{
                                                              'id' => 'asm::volume::equallogic',
                                                              'parameters' => [
                                                                  {'id' => 'title', 'value' => 'vol1'}
                                                              ]
                                                          }]
                                         },
                                         {'id' => 'k1',
                                          'puppetCertName' => 'k1',
                                          'asmGUID' => 'equallogic-1.1.1.1',
                                          'componentID'=>'s1',
                                          'resources' => [{
                                                              'id' => 'asm::volume::equallogic',
                                                              'parameters' => [
                                                                  {'id' => 'title', 'value' => 'vol2'}
                                                              ]
                                                          }]
                                         }
                                     ])
          ASM::Util.expects(:find_equallogic_iscsi_ip).with('equallogic-1.1.1.1').returns('127.0.1.1')
          ASM::Processor::Server.expects(:munge_hyperv_server).with(
              'bladeserver-serialno',
              {'asm::server' => {
                  'bladeserver-serialno' => {
                      'os_host_name' => 'foo',
                      'broker_type' => 'noop',
                      'os_image_version' => 'hyperv',
                      'serial_number' => 'SERIALNO',
                      'policy_name' => 'policy-foo-8000',
                      'razor_api_options' => {'url' => 'http://localhost:8080/api'},
                      'installer_options' => {
                          'os_type' => 'hyperv',
                          'agent_certname' => 'agent-foo'}}}},
              '127.0.1.1',
              ['vol1', 'vol2'],
              @sd.logger,
              false,
              'iscsi',
              'equallogic',
              'Fabric A'
          ).returns({})
          ASM::DeviceManagement.stubs(:parse_device_config).with(component['puppetCertName']).returns(
              {
                  :host => '172.17.5.170'
              }
          )
          @sd.process_server(component)
          @sd.debug = false
        end

      end

    end

    describe 'for boot from san server deployment' do
      it 'should configure a server' do

        @sd.stubs(:iterate_file).with(@o_file).returns(@o_file)
        @sd.stubs(:iterate_file).with(@r_file).returns(@r_file)

        ASM::Util.expects(:run_command_streaming).with(
            "sudo -i puppet asm process_node --filename #{@r_file} --run_type apply --always-override server-cert", "#{@o_file}") do |cmd|
          File.open(@o_file, 'w') do |fh|
            fh.write('Results: For 0 resources. 0 from our run failed. 0 not from our run failed. 0 updated successfully.')
          end
        end

        @data['serviceTemplate']['components'][0]['type'] = 'SERVER'
        @data['serviceTemplate']['components'][0]['resources'].push(
            {'id' => 'asm::server', 'parameters' => [
                {'id' => 'title', 'value' => 'foo'},
                {'id' => 'admin_password', 'value' => 'foo'},
                {'id' => 'os_host_name', 'value' => 'foo'},
                {'id' => 'os_image_type', 'value' => 'foo'}
            ]},
            {'id' => 'asm::idrac', 'parameters' => [
                {'id' => 'target_boot_device', 'value' => 'iSCSI'},
                {'id' => 'title', 'value' => 'foo'},
                {'id' => 'migrate_on_failure', 'value' => 'foo'},
                {'id' => 'attempted_servers', 'value' => 'foo'},
            ]},
            {'id' => 'asm::bios', 'parameters' => [
                {'id' => 'title', 'value' => 'foo'},
            ]}
        )

        @storage_component = [
            {'id' => 'k1',
             'puppetCertName' => 'k1',
             'asmGUID' => 'equallogic-1.1.1.1',
             'componentID'=>'s1',
             'resources' => [{
                                 'id' => 'asm::volume::equallogic',
                                 'parameters' => [
                                     {'id' => 'title', 'value' => 'vol1'}
                                 ]
                             }]
            },
            {'id' => 'k1',
             'puppetCertName' => 'k1',
             'asmGUID' => 'equallogic-1.1.1.1',
             'componentID'=>'s1',
             'resources' => [{
                                 'id' => 'asm::volume::equallogic',
                                 'parameters' => [
                                     {'id' => 'title', 'value' => 'vol2'}
                                 ]
                             }]
            }
        ]
        @sd.set_components_by_type('STORAGE',@storage_component )
        @sd.stubs(:find_related_components).with('STORAGE',@data['serviceTemplate']['components'][0]).returns(@storage_component)


        ASM::Util.stubs(:fetch_server_inventory).returns({'serviceTag' => 'server-cert', 'model' => 'M630'})
        ASM::DeviceManagement.stubs(:parse_device_config).with('k1').returns({})
        ASM::DeviceManagement.stubs(:run_puppet_device!).returns({})
        ASM::Util.stubs(:find_equallogic_iscsi_volume).returns('volname')
        ASM::Util.stubs(:find_equallogic_iscsi_ip).returns('127.0.0.1')

        @sd.stubs(:is_fcoe_enabled).returns(false)
        ASM::Util.stubs(:dell_cert?).returns(true)
        @sd.stubs(:process_tor_switches).returns(true)
        @sd.stubs(:process_san_switches).returns(true)
        @sd.stubs(:server_already_deployed).returns(false)
        ASM::DeviceManagement.stubs(:parse_device_config).with('server-cert').returns( {:host => '172.17.5.170' })
        ASM::DeviceManagement.expects(:puppet_run_success?).with("server-cert", 0, instance_of(Time), @o_file, @summary_file).returns([true, "x"])
        @sd.process(@data)
      end
    end

    describe 'for configuration processing' do
      describe 'when doing a teardown' do
        before(:each) do
          deployment_file = File.join(File.dirname(__FILE__), '..', '..', 'fixtures', 'service_deployment_teardown_test.json')
          @service_deployment = JSON.parse(File.read(deployment_file), :max_nesting => 100)
          @sd.stubs(:process_generic)
        end

        it 'should support teardown deploys' do
          @sd.is_teardown = true
          @sd.expects(:process_service_with_rules).with(@data)
          @sd.process(@data)
        end
      end

      it 'should set asm_hostname parameter if empty' do
        ASM::Util.expects(:run_command_streaming).with do |cmd, out_path|
          cmd.should == "sudo puppet asm process_node --debug --trace --filename #{@r_file} --run_type apply --statedir #{@r_dir} --always-override server-cert"
          out_path.should == "#{@o_file}"
          File.open(@o_file, 'w') do |fh|
            fh.write('Results: For 0 resources. 0 from our run failed. 0 not from our run failed. 0 updated successfully.')
          end
        end

        ASM::DeviceManagement.expects(:puppet_run_success?).with("server-cert", 0, instance_of(Time), @o_file, @summary_file).returns([true, "x"])

        ASM::Util.stubs(:get_preferred_ip).returns('172.0.0.1')
        @sd.stubs(:server_already_deployed).returns(false)
        mock_device_conf = {:provider => 'idrac'}
        ASM::DeviceManagement.stubs(:parse_device_config).returns(Hashie::Mash.new(mock_device_conf))
        @data['serviceTemplate']['components'][0]['type'] = 'CONFIGURATION'
        @data['serviceTemplate']['components'][0]['resources'].push(
            {'id' => 'asm::firmware', 'parameters' => [
                {'id' => 'title', 'value' => 'foo'},
                {'id' => 'asm_hostname', 'value' => nil},
                {'id' => 'other_param', 'value' => 'bar'}
            ]}
        )
        @sd.process(@data)
        resource_hash = YAML.load_file(@r_file)
        # Check asm_hostname has been substituted
        resource_hash['asm::firmware']['foo']['asm_hostname'].should == '172.0.0.1'
        # ... but others have not
        resource_hash['asm::firmware']['foo']['other_param'].should == 'bar'
      end

      it 'should not set asm_hostname parameter if provided' do
        ASM::Util.expects(:run_command_streaming).with do |cmd, out_path|
          cmd.should == "sudo puppet asm process_node --debug --trace --filename #{@r_file} --run_type apply --statedir #{@r_dir} --always-override server-cert"
          out_path.should == "#{@o_file}"
          File.open(@o_file, 'w') do |fh|
            fh.write('Results: For 0 resources. 0 from our run failed. 0 not from our run failed. 0 updated successfully.')
          end
        end

        ASM::Util.stubs(:get_preferred_ip).returns('172.0.0.1')
        mock_device_conf = {:provider => 'idrac'}
        ASM::DeviceManagement.stubs(:parse_device_config).returns(Hashie::Mash.new(mock_device_conf))
        ASM::DeviceManagement.expects(:puppet_run_success?).with("server-cert", 0, instance_of(Time), @o_file, @summary_file).returns([true, "x"])
        @data['serviceTemplate']['components'][0]['type'] = 'CONFIGURATION'
        @data['serviceTemplate']['components'][0]['resources'].push(
            {'id' => 'asm::firmware', 'parameters' => [
                {'id' => 'title', 'value' => 'foo'},
                {'id' => 'asm_hostname', 'value' => '192.168.1.1'},
                {'id' => 'other_param', 'value' => 'bar'}
            ]}
        )
        @sd.process(@data)
        YAML.load_file(@r_file)['asm::firmware']['foo']['asm_hostname'].should == '192.168.1.1'
      end

      it 'should use puppet apply for non-device types' do
        ASM::Util.expects(:run_command_streaming).with do |cmd, out_path|
          cmd.should == "sudo puppet asm process_node --debug --trace --filename #{@r_file} --run_type device --statedir #{@r_dir} --always-override server-cert"
          out_path.should == "#{@o_file}"
          File.open(@o_file, 'w') do |fh|
            fh.write('Results: For 0 resources. 0 from our run failed. 0 not from our run failed. 0 updated successfully.')
          end
        end

        ASM::Util.stubs(:get_preferred_ip).returns('172.0.0.1')
        mock_device_conf = {:provider => 'equallogic'} # should be puppet device
        ASM::DeviceManagement.stubs(:parse_device_config).returns(Hashie::Mash.new(mock_device_conf))
        ASM::DeviceManagement.stubs(:run_puppet_device!)
        ASM::DeviceManagement.expects(:puppet_run_success?).with("server-cert", 0, instance_of(Time), @o_file, @summary_file).returns([true, "x"])
        @data['serviceTemplate']['components'][0]['type'] = 'CONFIGURATION'
        @data['serviceTemplate']['components'][0]['resources'].push(
            {'id' => 'asm::firmware', 'parameters' => [
                {'id' => 'title', 'value' => 'foo'},
                {'id' => 'asm_hostname', 'value' => '192.168.1.1'},
            ]}
        )
        @sd.process(@data)
      end

    end

    it 'should report fcoe disabled for minimal server component' do
      comp = {
          'title' => 'cert',
          'resources' => [{
                              # in minimal server component this resource exists
                              # but without a network_configuration parameter
                              'id' => 'asm::esxiscsiconfig',
                              'parameters' => [{
                                                   'id' => 'title',
                                                   'value' => 'my-title'
                                               }],
                          }]}
      @sd.is_fcoe_enabled(comp).should be_nil
    end

  end

  describe 'when data is invalid' do

    it 'should warn when no serviceTemplate is defined' do
      @mock_log = mock('foo')
      @sd.expects(:logger).at_least_once.returns(@mock_log)
      @mock_log.stubs(:debug)
      @mock_log.expects(:info).with('Status: Started')
      @mock_log.expects(:info).with('Starting deployment ')
      @mock_log.expects(:warn).with('Service deployment data has no serviceTemplate defined')
      @mock_log.expects(:info).with('Status: Completed')
      @sd.process({})
    end

    it 'should warn when there are no components' do
      @mock_log = mock('foo')
      @sd.expects(:logger).at_least_once.returns(@mock_log)
      @mock_log.stubs(:debug)
      @mock_log.expects(:info).with('Status: Started')
      @mock_log.expects(:info).with('Starting deployment ')
      @mock_log.expects(:warn).with('service deployment data has no components')
      @mock_log.expects(:info).with('Status: Completed')
      @sd.process({'serviceTemplate' => {}})
    end

    it 'should fail when resources do not have types' do
      expect do
        @sd.process({'serviceTemplate' => {'components' => [
          {'id' => 'id', 'puppetCertName' => 'server-cert', 'type' => 'TEST', 'resources' => [
            {}
          ]}
        ]}})
      end.to raise_error('resource found with no type')
    end

    it 'should fail when resources do not have paremeters' do
      expect do
        @sd.process({'serviceTemplate' => {'components' => [
          {'type' => 'TEST', 'id' => 'id2','puppetCertName' => 'cert2', 'resources' => [
            {'id' => 'user'}
          ]}
        ]}})
      end.to raise_error('resource of type user has no parameters')
    end

    it 'should fail when component has no certname' do
      expect do
        @sd.process({'serviceTemplate' => {'components' => [
          {'type' => 'TEST', 'resources' => [
            {'id' => 'user'}
          ]}
        ]}})
      end.to raise_error('Component has no certname')
    end

    it 'should fail when a resource has no title' do
      expect do
        @sd.process({'serviceTemplate' => {'components' => [
          {'type' => 'TEST', 'id' => 'foo', 'puppetCertName' => 'cert4', 'resources' => [
            {'id' => 'user', 'parameters' => []}
          ]}
        ]}})
      end.to raise_error('Component has resource user with no title')
    end

  end

  describe 'dealing with duplicate certs in the same deployment' do
    before do
      @counter_files = [File.join(@tmp_dir, 'existing_file.yaml')]
    end
    after do
      @counter_files.each do |f|
        File.delete(f) if File.exists?(f)
      end
    end
    def write_counter_files
      @counter_files.each do |f|
        File.open(f, 'w') do |fh|
          fh.write('stuff')
        end
      end
    end
    it 'should be able to create file counters labeled 2 when files exist' do
      write_counter_files
      @sd.iterate_file(@counter_files.first).should == File.join(@tmp_dir, 'existing_file___2.yaml')
    end
    it 'should increment existing counter files' do
      @counter_files.push(File.join(@tmp_dir, 'existing_file___4.yaml'))
      write_counter_files
      @sd.iterate_file(@counter_files.first).should == File.join(@tmp_dir, 'existing_file___5.yaml')
    end
    it 'should return passed in file when no file exists' do
      @sd.iterate_file(@counter_files.first).should == File.join(@tmp_dir, 'existing_file.yaml')
    end
  end

  describe 'when checking agent status' do
    before do
      RestClient.stubs(:get)
        .with(URI.escape("http://localhost:7080/v3/nodes?query=[\"and\", [\"=\", [\"node\", \"active\"], true], [\"=\", \"name\", \"host\"]]]"),
        {:content_type => :json, :accept => :json})
          .returns('[{"name":"host"}]')

      RestClient.stubs(:get)
        .with(URI.escape("http://localhost:7080/v3/reports?query=[\"=\", \"certname\", \"host\"]&order-by=[{\"field\": \"receive-time\", \"order\": \"desc\"}]&limit=1"),
        {:content_type => :json, :accept => :json})
          .returns('[{"receive-time":"1970-01-01 01:00:00 +0000", "hash":"fooreport"}]')
    end

    it 'should be able to detect when a node has checked in' do
      sd = ASM::ServiceDeployment.new('123', @deployment_db)
      time = Time.at(0)
      cert_name = "host"
      sd.stubs(:puppetdb).returns(puppetdb)
      sd.await_agent_run_completion(cert_name, time, 0.1).should be(true)
    end

    it 'should raise PuppetEventException if node has not checked in' do
      sd = ASM::ServiceDeployment.new('123', @deployment_db)
      puppetdb.stubs(:successful_report_after?).returns(false)
      sd.stubs(:puppetdb).returns(puppetdb)
      expect{sd.await_agent_run_completion('host', Time.at(0), 1)}.to raise_exception(ASM::ServiceDeployment::PuppetEventException)
    end
  end

  describe 'when checking find related components' do
    before do
      sample_file = File.join(File.dirname(__FILE__), '..', '..',
                              'fixtures', 'find_related_components.json')
      data = JSON.parse(File.read(sample_file))
      comp_by_type = @sd.components_by_type(data)
      @sd.set_components_by_type('CLUSTER',  comp_by_type['CLUSTER'] )
      @sd.set_components_by_type('VIRTUALMACHINE',  comp_by_type['VIRTUALMACHINE'] )
      @components = data['serviceTemplate']['components']
    end
    it 'should return related component based on componentID' do
      expected = [{"id" => "ID1",
                   "componentID" => "COMPID1",
                   "type" => "CLUSTER",
                   "relatedComponents" => {"ID1" => "Virtual Machine 1"},
                   "resources" => {
                       "id" => "asm::cluster",
                       "parameters" => [{"id" => "datacenter"}]}
                  }]
      @sd.find_related_components('CLUSTER', @components[0]).should == expected
    end
    it 'should fail to related component based on ID' do
       @sd.find_related_components('VIRTUALMACHINE', @components[1]).should == []
    end
  end

  describe 'verifying service deployer internal configuration' do
    it 'configures directory' do
      @sd.stubs(:create_dir)
      @sd.send(:deployment_dir).should == File.join(@tmp_dir, @sd.id)
      @sd.send(:resources_dir).should == File.join(@tmp_dir, @sd.id, 'resources')
    end
  end

  describe 'when getting all switches' do
    it 'should find them' do
      certs = ['dell_ftos-1','server2', 'dell_ftos-2', 'brocade_foo-bar',
        'dell_iom-3', 'brocade_bar-4', 'dell_powerconnect-5',
        'dell_powerconnect-6' ]
      ASM::Util.stubs(:get_puppet_certs).returns(certs)
      ASM::Util.stubs(:fetch_managed_inventory).returns(certs.map do |cert|
        { 'refId' => cert, 'deviceType' => 'dellswitch' }
      end)
      ASM::Util.stubs(:get_dellswitch_fabric_mode).returns('Fabric-Services')
      ASM::Util.stubs(:get_cisconexus_features).returns('npv')
      puppetdb.stubs(:facts).returns({'model' => 'MXL'})
      ASM::Util.get_puppet_certs.should == certs
      @sd.get_all_switches
      @sd.configured_rack_switches.should == [ 'dell_ftos-1', 'dell_ftos-2',
        'dell_powerconnect-5', 'dell_powerconnect-6' ]
      @sd.configured_blade_switches.should == [ 'dell_ftos-1', 'dell_ftos-2','dell_iom-3' ]
      @sd.configured_brocade_san_switches.should == [ 'brocade_foo-bar', 'brocade_bar-4' ]
      ASM::Util.unstub(:get_puppet_certs)
    end
  end

  describe 'should find correct storage HBAs' do

    before do
      @mock_log = mock('logger')
      @mock_log.stub_everything
      @sd.stubs(:logger).returns(@mock_log)
      @endpoint = Hashie::Mash.new({:host => '127.0.0.1', :user => 'root', :password => 'password'})
      @hba_macs = {
          'vmhba33' => '00:10:18:C3:D9:7C',
          'vmhba34' => '00:10:18:C3:D9:7D',
          'vmhba35' => '00:10:18:C3:D9:7E',
          'vmhba36' => '00:10:18:C3:D9:7F',
      }
      @hbas = @hba_macs.keys.collect do |hba|
        {'Adapter' => hba, 'Description' => 'Broadcom iSCSI Adapter'}
      end
      ASM::Util.stubs(:esxcli).with(%w(iscsi adapter list), @endpoint, @mock_log).returns(@hbas)
      @hba_macs.each do |hba, mac|
        cmd = %w(iscsi adapter get --adapter).push(hba)
        serial = mac.gsub(/:/, '').downcase
        response = <<EOT
#{hba}
   Name: iqn.1998-01.com.vmware:host1051def.:674130862:35
   Alias: bnx2i-001018c3d97c
   Vendor: VMware
   Model: Broadcom iSCSI Adapter
   Description: Broadcom iSCSI Adapter
   Serial Number: #{serial}
   Hardware Version:
   Asic Version:
   Firmware Version:
   Option Rom Version:
   Driver Name: bnx2i-vmnic22
   Driver Version:
   TCP Protocol Supported: false
   Bidirectional Transfers Supported: false
   Maximum Cdb Length: 64
   Can Be NIC: true
   Is NIC: true
   Is Initiator: true
   Is Target: false
   Using TCP Offload Engine: true
   Using ISCSI Offload Engine: true

EOT
        ASM::Util.stubs(:esxcli).with(cmd, @endpoint, @mock_log, true).returns(response)
      end
    end

    it 'should fail if no hbas found' do
      expect do
        ASM::Util.stubs(:esxcli).with(%w(iscsi adapter list), @endpoint, @mock_log).returns([])
        @sd.parse_hbas(@endpoint, nil)
      end.to raise_error
    end

    it 'should return first two hbas when no iscsi macs specified' do
      ASM::Util.stubs(:esxcli).with(%w(iscsi adapter list), @endpoint, @mock_log).returns(@hbas)
      hbas = @sd.parse_hbas(@endpoint, nil)
      hbas.should == %w(vmhba33 vmhba34)
    end

    it 'should match hbas to iscsi macs' do
      ASM::Util.stubs(:esxcli).with(%w(iscsi adapter list), @endpoint, @mock_log).returns(@hbas)
      expected_hbas = %w(vmhba35 vmhba36)
      iscsi_macs = expected_hbas.collect { |hba| @hba_macs[hba] }
      hbas = @sd.parse_hbas(@endpoint, iscsi_macs)
      hbas.should == expected_hbas
    end

  end

  describe 'should find network partition information' do

    it 'should return storage network' do
      network_config = mock('network_configuration')
      network_config.stubs(:get_partitions).with('HYPERVISOR_MANAGEMENT').returns([])
      network_config.stubs(:get_partitions).with('HYPERVISOR_MIGRATION').returns([])
      network_config.stubs(:get_partitions).with('PRIVATE_LAN', 'PUBLIC_LAN').returns([])
      iscsi_networks = [Hashie::Mash.new({:type => 'STORAGE_ISCSI_SAN'})]
      network_config.stubs(:get_partitions).
          with('STORAGE_ISCSI_SAN', 'FILESHARE').
          returns([Hashie::Mash.new({:networkObjects => iscsi_networks})])

      info = @sd.gather_vswitch_info(network_config)
      info[:storage][:networks].should == iscsi_networks
    end

  end

  describe 'resolving hostnames' do

    before do
      @static = {'primaryDns' => '192.168.0.1',
                 'dnsSuffix' => 'foo.com',
                 'ipAddress' => '192.168.0.100'}
    end

    it 'should return the fqdn when hostname matches the IP' do
      hostname = 'host100'
      fqdn = "#{hostname}.#{@static['dnsSuffix']}"
      resolver = mock('resolver')
      resolved_ip = Resolv::IPv4.create(@static['ipAddress'])
      resolver.expects(:getaddress).with(hostname).returns(resolved_ip)
      spec = {:nameserver => [@static['primaryDns']], :search => Array(@static['dnsSuffix']), :ndots => 1}
      Resolv::DNS.expects(:open).with(spec).yields(resolver).returns(resolved_ip)
      @sd.lookup_hostname(hostname, @static).should == fqdn
    end

    it 'should return the fqdn when hostname matches the IP (both dns servers)' do
      @static['secondaryDns'] = '192.168.0.2'
      hostname = 'host100'
      fqdn = "#{hostname}.#{@static['dnsSuffix']}"
      resolver = mock('resolver')
      resolved_ip = Resolv::IPv4.create(@static['ipAddress'])
      resolver.expects(:getaddress).with(hostname).returns(resolved_ip)
      spec = {:nameserver => [@static['primaryDns'], @static['secondaryDns']],
              :search => Array(@static['dnsSuffix']), :ndots => 1}
      Resolv::DNS.expects(:open).with(spec).yields(resolver).returns(resolved_ip)
      @sd.lookup_hostname(hostname, @static).should == fqdn
    end

    it 'should return the fqdn when hostname matches the IP (secondary dns only)' do
      @static['primaryDns'] = nil
      @static['secondaryDns'] = '192.168.0.2'
      hostname = 'host100'
      fqdn = "#{hostname}.#{@static['dnsSuffix']}"
      resolver = mock('resolver')
      resolved_ip = Resolv::IPv4.create(@static['ipAddress'])
      resolver.expects(:getaddress).with(hostname).returns(resolved_ip)
      spec = {:nameserver => [@static['secondaryDns']],
              :search => Array(@static['dnsSuffix']), :ndots => 1}
      Resolv::DNS.expects(:open).with(spec).yields(resolver).returns(resolved_ip)
      @sd.lookup_hostname(hostname, @static).should == fqdn
    end

    it 'should return the hostname when it matches the IP without domain' do
      @static['dnsSuffix'] = nil
      hostname = 'host100'
      resolver = mock('resolver')
      resolved_ip = Resolv::IPv4.create(@static['ipAddress'])
      resolver.expects(:getaddress).with(hostname).returns(resolved_ip)
      spec = {:nameserver => [@static['primaryDns']], :search => Array(@static['dnsSuffix']), :ndots => 1}
      Resolv::DNS.expects(:open).with(spec).yields(resolver).returns(resolved_ip)
      @sd.lookup_hostname(hostname, @static).should == hostname
    end

    it 'should return the IP when host does not resolve' do
      hostname = 'host100'
      fqdn = "#{hostname}.#{@static['dnsSuffix']}"
      resolver = mock('resolver')
      resolver.expects(:getaddress).with(hostname).raises(Resolv::ResolvError.new)
      spec = {:nameserver => [@static['primaryDns']], :search => Array(@static['dnsSuffix']), :ndots => 1}
      Resolv::DNS.expects(:open).with(spec).yields(resolver)
      @sd.lookup_hostname(hostname, @static).should == @static['ipAddress']
    end

    it 'should return the IP when host resolves to a different IP' do
      hostname = 'host100'
      fqdn = "#{hostname}.#{@static['dnsSuffix']}"
      resolver = mock('resolver')
      resolved_ip = Resolv::IPv4.create('192.168.0.200')
      resolver.expects(:getaddress).with(hostname).returns(resolved_ip)
      spec = {:nameserver => [@static['primaryDns']], :search => Array(@static['dnsSuffix']), :ndots => 1}
      Resolv::DNS.expects(:open).with(spec).yields(resolver).returns(resolved_ip)
      @sd.lookup_hostname(hostname, @static).should == @static['ipAddress']
    end

    it "should not append the suffix when a fqdn is given" do
      hostname = "host100.my.com"
      fqdn = "host100.my.com"
      resolver = mock('resolver')
      resolved_ip = Resolv::IPv4.create('192.168.0.200')
      resolver.expects(:getaddress).with(fqdn).returns(resolved_ip)
      spec = {:nameserver => [@static['primaryDns']], :search => Array(@static['dnsSuffix']), :ndots => 1}
      Resolv::DNS.expects(:open).with(spec).yields(resolver).returns(resolved_ip)
      @sd.lookup_hostname(hostname, @static).should == @static['ipAddress']
    end
  end

end
