require 'spec_helper'
require 'asm/type'
require 'asm/type/server'

describe ASM::Type::Base do
  let(:logger) { stub(:debug => nil, :warn => nil, :info => nil) }

  let(:provider_name) { "rspec" }
  let(:provider_config) { {} }
  let(:provider_instance) { stub }
  let(:deployment) { stub(:debug? => false) }

  let(:service) { SpecHelper.service_from_fixture("Teardown_NetApp_VMware_Cluster_Server_VM.json") }
  let(:component) { service.components[1] }

  let(:type) { ASM::Type::Base.new(component, provider_name, provider_config, logger) }
  let(:provider) { stub }

  before :each do
    type.stubs(:provider).returns(provider)
    type.deployment = deployment
  end

  it "should set up the forwardable module" do
    expect(ASM::Type::Base).to respond_to(:def_delegators)
  end

  describe "#service_teardown?" do
    it "should report the service teardown via the component" do
      service.expects(:teardown?)
      type.service_teardown?
    end
  end

  describe "#process_generic" do
    it "should return the result from process_generic" do
      deployment.expects(:process_generic).returns({"rspec" => 1})
      expect(type.process_generic("x", {}, "apply")).to eq({"rspec" => 1})
    end
  end

  describe "#update_inventory" do
    it "should update the inventory if it should" do
      provider.expects(:should_inventory?).returns(true)
      provider.expects(:update_inventory)
      type.update_inventory
    end

    it "should not when it shouldnt" do
      provider.expects(:should_inventory?).returns(false)
      provider.expects(:update_inventory).never
      type.update_inventory
    end
  end

  describe "#supports_resource?" do
    it "should support volumes" do
      provider.expects(:volume_supported?)
      type.supports_resource?(stub(:type_name => "volume", :provider_path => "rspec/rspec"))
    end

    it "should support servers" do
      provider.expects(:server_supported?)
      type.supports_resource?(stub(:type_name => "server", :provider_path => "rspec/rspec"))
    end

    it "should support clusters" do
      provider.expects(:cluster_supported?)
      type.supports_resource?(stub(:type_name => "cluster", :provider_path => "rspec/rspec"))
    end

    it "should support virtualmachines" do
      provider.expects(:virtualmachine_supported?)
      type.supports_resource?(stub(:type_name => "virtualmachine", :provider_path => "rspec/rspec"))
    end

    it "should support controllers" do
      provider.expects(:controller_supported?)
      type.supports_resource?(stub(:type_name => "controller", :provider_path => "rspec/rspec"))
    end
  end

  describe "#provider_path" do
    it "should give the correct path" do
      expect(type.provider_path).to eq("base/rspec")
    end
  end

  describe "#initialize" do
    it "should set the puppet_certname from the service component" do
      expect(type.puppet_certname).to eq(component.puppet_certname)
    end
  end

  describe "#get_provider_class" do
    it "should be able to fetch provider class constants" do
      provider_klass = mock
      ASM::Provider::Base.expects(:const_get).with("Rspec").returns(provider_klass)

      type.get_provider_class("rspec")
    end
  end

  describe "#delegate" do
    it "should delegate to the right object and log it" do
      (object = stub).expects(:delegated_call).with(1,2,3,4).once
      object.stubs(:to_s).returns("#<rspec mock>")
      logger.expects(:debug).with(regexp_matches(/calling delegated method delegated_call on #<rspec mock>/))

      type.delegate(object, :delegated_call, 1, 2, 3, 4)
    end
  end

  describe "#cert2serial" do
    it "should convert the supplied certname" do
      expect(type.cert2serial("rackserver-rspec")).to eq("RSPEC")
    end

    it "should default to the server certname" do
      expect(type.cert2serial).to eq("CDQSGT1")
    end
  end

  describe "#related_components" do
    it "should find all related servers" do
      type = ASM::Type::Base.new(service.components[0], provider_name, provider_config, logger)

      servers = type.related_components("SERVER")
      expect(servers.size).to eq(2)

      expect(servers[0].puppet_certname).to eq("bladeserver-5qykcy1")
      expect(servers[1].puppet_certname).to eq("bladeserver-cdqsgt1")
    end
  end

  describe "#related_volumes" do
    it "should find all volumes" do
      type.expects(:related_components).with("STORAGE")
      type.related_volumes
    end
  end

  describe "#related_clusters" do
    it "should find all clusters" do
      type.expects(:related_components).with("CLUSTER")
      type.related_clusters
    end
  end

  describe "#related_servers" do
    it "should find all servers" do
      type.expects(:related_components).with("SERVER")
      type.related_servers
    end
  end

  describe "#related_cluster" do
    it "should return the first found cluster" do
      type.expects(:related_clusters).returns(["rspec0", "rspec1"])
      expect(type.related_cluster).to eq("rspec0")
    end
  end

  describe "#related_server" do
    it "should return the first found server" do
      type.expects(:related_servers).returns(["rspec0", "rspec1"])
      expect(type.related_server).to eq("rspec0")
    end
  end
end
