require 'spec_helper'
require 'asm/type'

ASM::Type.load_providers!

describe ASM::Type::Server do
  let(:service) { SpecHelper.service_from_fixture("Teardown_EQL_VMware_Cluster.json") }
  let(:logger) { stub(:debug => nil, :warn => nil, :info => nil) }
  let(:db) { stub }
  let(:deployment) { stub(:id => "1234", :db => db) }
  let(:server_component) { service.components_by_type("SERVER").first }

  let(:provider) { stub(:hostname => "rspec.host") }
  let(:type) { server_component.to_resource(deployment, logger) }
  let(:raw_network_params) { server_component.configuration["asm::esxiscsiconfig"]["bladeserver-15kvd42"] }

  before(:each) do
    type.stubs(:provider).returns(provider)
  end

  describe "#deployment_completed?" do
    it "should return true if the db says its done" do
      type.expects(:db_execution_status).returns("complete")
      expect(type.deployment_completed?).to be(true)
    end

    it "should return if not completed in the db and its a bfs machine" do
      type.expects(:db_execution_status).returns(nil)
      type.expects(:boot_from_san?).returns(true)
      type.expects(:razor_status).never
      expect(type.deployment_completed?).to be(false)
    end

    it "should check razor otherwise" do
      type.stubs(:db_execution_status).returns(nil)
      type.stubs(:boot_from_san?).returns(false)
      type.stubs(:has_os?).returns(true)

      type.expects(:razor_status).returns("rspec")
      expect(type.deployment_completed?).to be(false)

      type.expects(:razor_status).returns("boot_local")
      expect(type.deployment_completed?).to be(true)

      type.expects(:razor_status).returns("boot_local_2")
      expect(type.deployment_completed?).to be(true)
    end
  end

  describe "#db_execution_status" do
    it "should return the found status otherwise nil" do
      db.expects(:get_component_status).with("B0151383-9478-459F-A420-CAED5DE5B063").returns({:status => "rspec"})
      expect(type.db_execution_status).to eq("rspec")

      db.expects(:get_component_status).with("B0151383-9478-459F-A420-CAED5DE5B063").returns(nil)
      expect(type.db_execution_status).to eq(nil)
    end
  end

  describe "#razor_status" do
    it "should report the razor status when the node is found" do
      ASM::Razor.expects(:new).returns(razor = stub)
      razor.expects(:find_node).with("15KVD42").returns({"name" => "rspec"})
      razor.expects(:task_status).with("rspec", "policy-rspec.host-1234").returns({:status => "rspec"})

      expect(type.razor_status).to eq({:status => "rspec"})
    end
  end

  describe "#rackserver?" do
    it "should detect rack servers correctly" do
      type.expects(:puppet_certname).returns("rackserver-foo")
      expect(type.rackserver?).to be(true)

      type.expects(:puppet_certname).returns("rspecserver-foo")
      expect(type.rackserver?).to be(false)
    end
  end
  describe "#fcoe?" do
    it "should report false when there are no fcoe networks" do
      type.expects(:fcoe_san_networks).returns({})
      expect(type.fcoe?).to be(false)
    end

    it "should report false when there are no fcoe networks" do
      type.expects(:fcoe_san_networks).returns({1=>2})
      expect(type.fcoe?).to be(true)
    end
  end

  describe "#fcoe_san_networks" do
    it "should retrieve the STORAGE_FCOE_SAN networks" do
      type.expects(:network_config).returns(nc = stub)
      nc.expects(:get_networks).with("STORAGE_FCOE_SAN").returns({"rspec" => 1})
      expect(type.fcoe_san_networks).to eq({"rspec" => 1})
    end
  end

  describe "#fabric_info" do
    let(:rack_network_config_json) { JSON.parse(SpecHelper.load_fixture("network_configuration/rack_partitioned.json")) }
    let(:rack_network_config) { ASM::NetworkConfiguration.new(rack_network_config_json) }

    it "should create correct rack server info based on cards" do
      type.stubs(:network_config).returns(rack_network_config)
      expect(type.fabric_info).to eq({"Fabric A" => 2})
    end

    it "should return all fabrics (even non-enabled ones) for non rack network configs" do
      expected = {}

      server_component.resource_by_id("asm::esxiscsiconfig")["network_configuration"]["fabrics"].each do |f|
        expected[f["name"]] = Integer(f["nictype"])
      end

      expect(expected).to_not be_empty
      expect(type.fabric_info).to eq(expected)
    end
  end

  describe "#network_params" do
    it "should fetch asm::esxiscsiconfig from the component" do
      expect(type.network_params).to eq(raw_network_params)
    end

    it "should return an empty hash when none is found" do
      server_component.expects(:resource_by_id).with("asm::esxiscsiconfig").returns(nil)
      expect(type.network_params).to eq({})
    end
  end

  describe "#network_config" do
    it "should create instances of ASM::NetworkConfiguration" do
      ASM::NetworkConfiguration.expects(:new).with(raw_network_params["network_configuration"]).returns(config = mock)
      expect(type.network_config).to eq(config)
    end
  end

  describe "#management_network" do
    it "should retrieve the management network from the network config" do
      config = ASM::NetworkConfiguration.new(raw_network_params["network_configuration"])

      expect(type.management_network).to eq(config.get_network('HYPERVISOR_MANAGEMENT'))
      expect(type.management_network.name).to eq("HypervisorManagement")
      expect(type.management_network.id).to eq("ff8080814dbf2d1d014dc298d147006f")
      expect(type.management_network.staticNetworkConfiguration["ipAddress"]).to eq("172.28.10.63")
    end
  end

  describe "#static_network_config" do
    it "should retrieve the management network static config" do
      expect(type.static_network_config).to eq(type.management_network["staticNetworkConfiguration"])
    end
  end

  describe "#primary_dnsserver" do
    it "should retrieve the primaryDns" do
      expect(type.primary_dnsserver).to eq("172.20.0.8")
    end
  end

  describe "#secondary_dnsserver" do
    it "should retrieve the secondaryDns" do
      type.stubs(:static_network_config).returns("secondaryDns" => "8.8.8.8")
      expect(type.secondary_dnsserver).to eq("8.8.8.8")
    end
  end

  describe "#hostip" do
    it "should supply the hosts ip in the static network" do
      expect(type.hostip).to eq(type.static_network_config["ipAddress"])
    end
  end

  describe "#hostname" do
    it "should get the hostname from the procider" do
      type.expects(:provider).returns(mock(:hostname))
      type.hostname
    end
  end

  describe "#hyperv_config" do
    it "should return the data from the provider" do
      provider.stubs(:to_hash).with(true, :hyperv).returns({"rspec" => true})
      expect(type.hyperv_config).to eq({"rspec" => true})
    end
  end

  [:delete_server_cert!, :leave_cluster!, :clean_related_volumes!, :clean_virtual_identities!, :power_off!, :delete_server_node_data!].each do |m|
    describe "#%s" % m do
      it "should delegate to the provider" do
        type.expects(:delegate).with(provider, m).once
        type.send(m)
      end
    end
  end
end
